#
# Copyright (c) 2015, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Module for managing the NTP configuration in EOS

This module provides an API for configuring NTP resources using
EOS and eAPI.

Arguments:
    name (string): The interface port that specifies the NTP source.
"""

import re

from pyeapi.api import Entity


class Ntp(Entity):
    """The Ntp class implements global NTP router configuration
    """

    def __init__(self, *args, **kwargs):
        super(Ntp, self).__init__(*args, **kwargs)

    def get(self):
        """Returns the current NTP configuration

        The Ntp resource returns the following:

            * source_interface (str): The interface port that specifies
                                      NTP server
            * servers (list): A list of the NTP servers that have been
                              assigned to the node. Each entry in the
                              list is a key/value pair of the name of
                              the server as the key and None or 'prefer'
                              as the value if the server is preferred.

        Returns:
            A Python dictionary object of key/value pairs that represents
            the current NTP configuration of the node::

                {
                    "source_interface": 'Loopback0',
                    'servers': [
                        { '1.1.1.1': None },
                        { '1.1.1.2': 'prefer' },
                        { '1.1.1.3': 'prefer' },
                        { '1.1.1.4': None },
                    ]
                }
        """
        config = self.config
        if not config:
            return None

        response = dict()
        response.update(self._parse_source_interface(config))
        response.update(self._parse_servers(config))

        return response

    def _parse_source_interface(self, config):
        if self.version_number >= '4.23':
            match = re.search(r'^ntp local-interface ([^\s]+)', config, re.M)
        else:
            match = re.search(r'^ntp source ([^\s]+)', config, re.M)
        value = match.group(1) if match else None
        return dict(source_interface=value)

    def _parse_servers(self, config):
        matches = re.findall(r'ntp server ([\S]+) ?(prefer)?', config, re.M)
        value = []
        for match in matches:
            server = match[0]
            prefer = match[1] if match[1] == 'prefer' else None
            value.append({server: prefer})
        return dict(servers=value)

    def create(self, name):
        """Instantiate the NTP by setting the source interface.

        Args:
            name (string): The interface port that specifies the NTP source.

        Returns:
            True if the operation succeeds, otherwise False.
        """
        return self.set_source_interface(name)

    def delete(self):
        """Delete the NTP source entry from the node.

        Returns:
            True if the operation succeeds, otherwise False.
        """
        if self.version_number >= '4.23':
            cmd = self.command_builder('ntp local-interface', disable=True)
        else:
            cmd = self.command_builder('ntp source', disable=True)
        return self.configure(cmd)

    def default(self):
        """Default the NTP source entry from the node.

        Returns:
            True if the operation succeeds, otherwise False.
        """
        if self.version_number >= '4.23':
            cmd = self.command_builder('ntp local-interface', default=True)
        else:
            cmd = self.command_builder('ntp source', default=True)
        return self.configure(cmd)

    def set_source_interface(self, name):
        """Assign the NTP source on the node

        Args:
            name (string): The interface port that specifies the NTP source.

        Returns:
            True if the operation succeeds, otherwise False.
        """
        if self.version_number >= '4.23':
            cmd = self.command_builder('ntp local-interface', value=name)
        else:
            cmd = self.command_builder('ntp source', value=name)
        return self.configure(cmd)

    def add_server(self, name, prefer=False):
        """Add or update an NTP server entry to the node config

        Args:
            name (string): The IP address or FQDN of the NTP server.
            prefer (bool): Sets the NTP server entry as preferred if True.

        Returns:
            True if the operation succeeds, otherwise False.
        """
        if not name or re.match(r'^[\s]+$', name):
            raise ValueError('ntp server name must be specified')
        if prefer:
            name = '%s prefer' % name
        cmd = self.command_builder('ntp server', value=name)
        return self.configure(cmd)

    def remove_server(self, name):
        """Remove an NTP server entry from the node config

        Args:
            name (string): The IP address or FQDN of the NTP server.

        Returns:
            True if the operation succeeds, otherwise False.
        """
        cmd = self.command_builder('no ntp server', value=name)
        return self.configure(cmd)

    def remove_all_servers(self):
        """Remove all NTP server entries from the node config

        Returns:
            True if the operation succeeds, otherwise False.
        """
        # 'no ntp' removes all server entries.
        # For command_builder, disable command 'ntp' gives the desired command
        cmd = self.command_builder('ntp', disable=True)
        return self.configure(cmd)


def instance(node):
    """Returns an instance of Ntp

    Args:
        node (Node): The node argument passes an instance of Node to the
            resource

    Returns:
        object: An instance of Ntp
    """
    return Ntp(node)



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""API Module for working with EOS local user resources

The Users resource provides configuration of local user resources for
an EOS node.

Parameters:

    username (string): The username parameter maps to the local username
        defined in the running-config.
    nopassword (boolean): Configures the username to use no password at login.
        This parameter is mutually exclusive with secret
    privilege (integer): Configures the user privilege level in EOS
    role (string): Configures the users role in EOS
    secret (string): Configures the users secret (password) to use at login.
        This parameter is mutually exclusive with secret and is used in
        conjunction with format.
    format (string): Configures the format of the secret value.  Accepted
        values for format are "cleartext", "md5", "nologin" and "sha512"

"""

import re

from pyeapi.api import EntityCollection

DEFAULT_ENCRYPTION = 'cleartext'
ENCRYPTION_MAP = {'cleartext': 0, 'md5': 5, 'sha512': 'sha512', 'nologin': '*'}


def isprivilege(value):
    """Checks value for valid privilege level

    Args:
        value (str, int): Checks if value is a valid user privilege

    Returns:
        True if the value is valid, otherwise False
    """
    try:
        value = int(value)
        return 0 <= value < 16
    except ValueError:
        return False


class Users(EntityCollection):
    """The Users class provides a configuration resource for local users.
    The regex used here parses the running configuration to find username
    entries. There is extra logic in the regular expression to store
    the username as 'user' and then creates a backreference to find a
    following configuration line that might contain the users sshkey.
    """

    def get(self, name):
        """Returns the local user configuration as a resource dict

        Args:
            name (str): The username to return from the nodes global running-
                config.

        Returns:
            dict: A resource dict object

            If the `name` does not exist, then None is returned
        """
        return self.getall().get(name)

    def getall(self):
        """Returns all local users configuration as a resource dict

        Returns:
            dict: A dict of usernames with a nested resource dict object
        """
        if self.version_number >= '4.23':
            self.users_re = re.compile(r'username (?P<user>[^\s]+) '
                                       r'privilege (\d+)'
                                       r'(?: role ([^\s]+))?'
                                       r'(?: (nopassword))?'
                                       r'(?: secret (0|5|7|sha512) (.+))?'
                                       r'.*$\n(?:username (?P=user) '
                                       r'ssh.key (.+)$)?', re.M)
        else:
            self.users_re = re.compile(r'username (?P<user>[^\s]+) '
                                       r'privilege (\d+)'
                                       r'(?: role ([^\s]+))?'
                                       r'(?: (nopassword))?'
                                       r'(?: secret (0|5|7|sha512) (.+))?'
                                       r'.*$\n(?:username (?P=user) '
                                       r'sshkey (.+)$)?', re.M)

        users = self.users_re.findall(self.config, re.M)
        resources = dict()
        for user in users:
            resources.update(self._parse_username(user))
        return resources

    def _parse_username(self, config):
        """Scans the config block and returns the username as a dict

        Args:
            config (str): The config block to parse

        Returns:
            dict: A resource dict that is intended to be merged into the
                user resource
        """
        (username, priv, role, nopass, fmt, secret, sshkey) = config
        resource = dict()
        resource['privilege'] = priv
        resource['role'] = role
        resource['nopassword'] = nopass == 'nopassword'
        resource['format'] = fmt
        resource['secret'] = secret
        if self.version_number >= '4.23':
            resource['ssh-key'] = sshkey
        else:
            resource['sshkey'] = sshkey
        return {username: resource}

    def create(self, name, nopassword=None, secret=None, encryption=None):
        """Creates a new user on the local system.

        Creating users requires either a secret (password) or the nopassword
        keyword to be specified.

        Args:
            name (str): The name of the user to craete

            nopassword (bool): Configures the user to be able to authenticate
                without a password challenage

            secret (str): The secret (password) to assign to this user

            encryption (str): Specifies how the secret is encoded.  Valid
                values are "cleartext", "md5", "nologin", "sha512".
                The default is "cleartext"

        Returns:
            True if the operation was successful otherwise False

        Raises:
            TypeError: if the required arguments are not satisfied
        """
        if secret is not None or encryption == 'nologin':
            return self.create_with_secret(name, secret, encryption)
        elif nopassword is True:
            return self.create_with_nopassword(name)
        else:
            raise TypeError('either "nopassword" or "secret" must be '
                            'specified to create a user')

    def create_with_secret(self, name, secret, encryption):
        """Creates a new user on the local node

        Args:
            name (str): The name of the user to craete

            secret (str): The secret (password) to assign to this user

            encryption (str): Specifies how the secret is encoded.  Valid
                values are "cleartext", "md5", "nologin" and "sha512".
                The default is "cleartext"

        Returns:
            True if the operation was successful otherwise False
        """
        try:
            encryption = encryption or DEFAULT_ENCRYPTION
            enc = ENCRYPTION_MAP[encryption]
        except KeyError:
            raise TypeError('encryption must be one of "cleartext", "md5"'
                            ' "nologin" or "sha512"')

        cmd = 'username %s secret %s %s' % (name, enc, secret)
        if encryption == 'nologin':
            cmd = 'username %s secret %s' % (name, enc)
        return self.configure(cmd)

    def create_with_nopassword(self, name):
        """Creates a new user on the local node

        Args:
            name (str): The name of the user to create

        Returns:
            True if the operation was successful otherwise False
        """
        return self.configure('username %s nopassword' % name)

    def delete(self, name):
        """Deletes the local username from the config

        Args:
            name (str): The name of the user to delete

        Returns:
            True if the operation was successful otherwise False
        """
        if name == 'admin':
            raise TypeError('the admin user cannot be deleted.')

        return self.configure('no username %s' % name)

    def default(self, name):
        """Configures the local username using the default keyword

        Args:
            name (str): The name of the user to configure

        Returns:
            True if the operation was successful otherwise False
        """
        return self.configure('default username %s' % name)

    def set_privilege(self, name, value=None):
        """Configures the user privilege value in EOS

        Args:
            name (str): The name of the user to craete

            value (int): The privilege value to assign to the user.  Valid
                values are in the range of 0 to 15

        Returns:
            True if the operation was successful otherwise False

        Raises:
            TypeError: if the value is not in the valid range
        """
        cmd = 'username %s' % name
        if value is not None:
            if not isprivilege(value):
                raise TypeError('priviledge value must be between 0 and 15')
            cmd += ' privilege %s' % value
        else:
            cmd += ' privilege 1'
        return self.configure(cmd)

    def set_role(self, name, value=None, default=False, disable=False):
        """Configures the user role vale in EOS

        Args:
            name (str): The name of the user to create

            value (str): The value to configure for the user role

            default (bool): Configure the user role using the EOS CLI
                default command

            disable (bool): Negate the user role using the EOS CLI no command

        Returns:
            True if the operation was successful otherwise False
        """
        cmd = self.command_builder('username %s role' % name, value=value,
                                   default=default, disable=disable)
        return self.configure(cmd)

    def set_sshkey(self, name, value=None, default=False, disable=False):
        """Configures the user sshkey

        Args:
            name (str): The name of the user to add the sshkey to

            value (str): The value to configure for the sshkey.

            default (bool): Configure the sshkey using the EOS CLI
                default command

            disable (bool): Negate the sshkey using the EOS CLI no command

        Returns:
            True if the operation was successful otherwise False
        """
        if self.version_number >= '4.23':
            cmd = self.command_builder('username %s ssh-key' % name,
                                       value=value,
                                       default=default, disable=disable)
        else:
            cmd = self.command_builder('username %s sshkey' % name,
                                       value=value,
                                       default=default, disable=disable)
        return self.configure(cmd)


def instance(node):
    """Returns an instance of Users

    This method will create and return an instance of the Users object passing
    the value of API to the object.  The instance method is required for the
    resource to be autoloaded by the Node object

    Args:
        node (Node): The node argument passes an instance of Node to the
            resource
    """
    return Users(node)



#
# Copyright (c) 2015, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Module for working with EOS routemap resources

The Routemap resource provides configuration management of global route-map
resources on an EOS node. It provides the following class implementations:

    * Routemaps - Configures routemaps in EOS

Notes:
    The set and match attributes produce a list of strings with The
    corresponding configuration. These strings will omit the preceeding
    set or match words, respectively.
"""

import re

from pyeapi.api import EntityCollection


class Routemaps(EntityCollection):
    """The Routemaps class provides management of the routemaps configuration

    The Routemaps class is derived from Entity and provides an API for working
    with the nodes routemaps configuraiton.
    """

    def get(self, name):
        """Provides a method to retrieve all routemap configuration
        related to the name attribute.

        Args:
            name (string): The name of the routemap.

        Returns:
            None if the specified routemap does not exists. If the routermap
            exists a dictionary will be provided as follows::

                {
                    'deny': {
                            30: {
                                    'continue': 200,
                                    'description': None,
                                    'match': ['as 2000',
                                              'source-protocol ospf',
                                              'interface Ethernet2'],
                                    'set': []
                                }
                            },
                    'permit': {
                            10: {
                                    'continue': 100,
                                    'description': None,
                                    'match': ['interface Ethernet1'],
                                    'set': ['tag 50']},
                            20: {
                                    'continue': 200,
                                    'description': None,
                                    'match': ['as 2000',
                                              'source-protocol ospf',
                                              'interface Ethernet2'],
                                    'set': []
                                }
                            }
                }
        """
        if not self.get_block(r'route-map\s%s\s\w+\s\d+' % name):
            return None

        return self._parse_entries(name)

    def getall(self):
        resources = dict()
        routemaps_re = re.compile(r'^route-map\s([\w-]+)\s\w+\s\d+$', re.M)
        for name in routemaps_re.findall(self.config):
            routemap = self.get(name)
            if routemap:
                resources[name] = routemap
        return resources

    def _parse_entries(self, name):

        routemap_re = re.compile(r'^route-map\s%s\s(\w+)\s(\d+)$'
                                 % name, re.M)
        entries = list()
        for entry in routemap_re.findall(self.config):
            resource = dict()
            action, seqno = entry
            routemap = self.get_block(r'route-map\s%s\s%s\s%s'
                                      % (name, action, seqno))

            resource = dict(name=name, action=action, seqno=seqno, attr=dict())
            resource['attr'].update(self._parse_match_statements(routemap))
            resource['attr'].update(self._parse_set_statements(routemap))
            resource['attr'].update(self._parse_continue_statement(routemap))
            resource['attr'].update(self._parse_description(routemap))
            entries.append(resource)

        return self._merge_entries(entries)

    def _merge_entries(self, entries):
        response = dict()
        for e in entries:
            action = e['action']
            seqno = int(e['seqno'])
            if not response.get(action):
                response[action] = dict()
            response[action][seqno] = e['attr']

        return response

    def _parse_match_statements(self, config):
        match_re = re.compile(r'^\s+match\s(.+)$', re.M)
        return dict(match=match_re.findall(config))

    def _parse_set_statements(self, config):
        set_re = re.compile(r'^\s+set\s(.+)$', re.M)
        return dict(set=set_re.findall(config))

    def _parse_continue_statement(self, config):
        continue_re = re.compile(r'^\s+continue\s(\d+)$', re.M)
        match = continue_re.search(config)
        value = int(match.group(1)) if match else None
        return {'continue': value}

    def _parse_description(self, config):
        desc_re = re.compile(r'^\s+description\s(.+)$', re.M)
        match = desc_re.search(config)
        value = match.group(1) if match else None
        return dict(description=value)

    def create(self, name, action, seqno):
        """Creates a new routemap on the node

        Note:
            This method will attempt to create the routemap regardless
            if the routemap exists or not.  If the routemap already exists
            then this method will still return True.

        Args:
            name (string): The full name of the routemap.
            action (string): The action to take for this routemap clause.
            seqno (integer): The sequence number for the routemap clause.

        Returns:
            True if the routemap could be created otherwise False (see Note)

        """
        return self.configure('route-map %s %s %s' % (name, action, seqno))

    def delete(self, name, action, seqno):
        """Deletes the routemap from the node

        Note:
            This method will attempt to delete the routemap from the nodes
            operational config.  If the routemap does not exist then this
            method will not perform any changes but still return True

        Args:
            name (string): The full name of the routemap.
            action (string): The action to take for this routemap clause.
            seqno (integer): The sequence number for the routemap clause.

        Returns:
            True if the routemap could be deleted otherwise False (see Node)

        """
        return self.configure('no route-map %s %s %s' % (name, action, seqno))

    def default(self, name, action, seqno):
        """Defaults the routemap on the node

        Note:
            This method will attempt to default the routemap from the nodes
            operational config. Since routemaps do not exist by default,
            the default action is essentially a negation and the result will
            be the removal of the routemap clause.
            If the routemap does not exist then this
            method will not perform any changes but still return True

        Args:
            name (string): The full name of the routemap.
            action (string): The action to take for this routemap clause.
            seqno (integer): The sequence number for the routemap clause.

        Returns:
            True if the routemap could be deleted otherwise False (see Node)

        """
        return self.configure('default route-map %s %s %s'
                              % (name, action, seqno))

    def set_match_statements(self, name, action, seqno, statements):
        """Configures the match statements within the routemap clause.
        The final configuration of match statements will reflect the list
        of statements passed into the statements attribute. This implies
        match statements found in the routemap that are not specified in the
        statements attribute will be removed.

        Args:
            name (string): The full name of the routemap.
            action (string): The action to take for this routemap clause.
            seqno (integer): The sequence number for the routemap clause.
            statements (list): A list of the match-related statements. Note
                               that the statements should omit the leading
                               match.

        Returns:
            True if the operation succeeds otherwise False
        """
        try:
            current_statements = self.get(name)[action][seqno]['match']
        except Exception:
            current_statements = []

        commands = list()

        # remove set statements from current routemap
        for entry in set(current_statements).difference(statements):
            commands.append('route-map %s %s %s' % (name, action, seqno))
            commands.append('no match %s' % entry)

        # add new set statements to the routemap
        for entry in set(statements).difference(current_statements):
            commands.append('route-map %s %s %s' % (name, action, seqno))
            commands.append('match %s' % entry)

        return self.configure(commands) if commands else True

    def set_set_statements(self, name, action, seqno, statements):
        """Configures the set statements within the routemap clause.
        The final configuration of set statements will reflect the list
        of statements passed into the statements attribute. This implies
        set statements found in the routemap that are not specified in the
        statements attribute will be removed.

        Args:
            name (string): The full name of the routemap.
            action (string): The action to take for this routemap clause.
            seqno (integer): The sequence number for the routemap clause.
            statements (list): A list of the set-related statements. Note that
                               the statements should omit the leading set.

        Returns:
            True if the operation succeeds otherwise False
        """
        try:
            current_statements = self.get(name)[action][seqno]['set']
        except Exception:
            current_statements = []

        commands = list()

        # remove set statements from current routemap
        for entry in set(current_statements).difference(statements):
            commands.append('route-map %s %s %s' % (name, action, seqno))
            commands.append('no set %s' % entry)

        # add new set statements to the routemap
        for entry in set(statements).difference(current_statements):
            commands.append('route-map %s %s %s' % (name, action, seqno))
            commands.append('set %s' % entry)

        return self.configure(commands) if commands else True

    def set_continue(self, name, action, seqno, value=None, default=False,
                     disable=False):
        """Configures the routemap continue value

        Args:
            name (string): The full name of the routemap.
            action (string): The action to take for this routemap clause.
            seqno (integer): The sequence number for the routemap clause.
            value (integer): The value to configure for the routemap continue
            default (bool): Specifies to default the routemap continue value
            disable (bool): Specifies to negate the routemap continue value

        Returns:
            True if the operation succeeds otherwise False is returned
        """
        commands = ['route-map %s %s %s' % (name, action, seqno)]
        if default:
            commands.append('default continue')
        elif disable:
            commands.append('no continue')
        else:
            if not str(value).isdigit() or value < 1:
                raise ValueError('seqno must be a positive integer unless '
                                 'default or disable is specified')
            commands.append('continue %s' % value)

        return self.configure(commands)

    def set_description(self, name, action, seqno, value=None, default=False,
                        disable=False):
        """Configures the routemap description

        Args:
            name (string): The full name of the routemap.
            action (string): The action to take for this routemap clause.
            seqno (integer): The sequence number for the routemap clause.
            value (string): The value to configure for the routemap description
            default (bool): Specifies to default the routemap description value
            disable (bool): Specifies to negate the routemap description

        Returns:
            True if the operation succeeds otherwise False is returned
        """
        commands = ['route-map %s %s %s' % (name, action, seqno)]
        if value is not None:
            # Before assigning a new description, clear any existing desc
            commands.append(self.command_builder('description', disable=True))
        commands.append(self.command_builder('description', value=value,
                                             default=default, disable=disable))
        return self.configure(commands)


def instance(node):
    """Returns an instance of Routemaps

    Args:
        node (Node): The node argument passes an instance of Node to the
            resource

    Returns:
        object: An instance of Routemaps
    """
    return Routemaps(node)



#
# Copyright (c) 2017, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Module for working with EOS VRF resources

The Vrfs resource provides configuration of VRF resources for an EOS
node.

Parameters:

    name (string): The name parameter maps to the VRF name in EOS.  Valid
        values include any consecutive sequence of numbers, letters and
        underscore up to the maximum number of characters.  This parameter
        is defaultable.
    description (string): The vrf description set by the user
    ipv4_routing (bool): Tells whether IPv4 routing is enabled on the VRF
    ipv6_routing (bool): Tells whether IPv6 unicast routing is enabled on the
        VRF

"""

import re

from pyeapi.api import EntityCollection
from pyeapi.utils import make_iterable

RD_RE = re.compile(r'(?:\srd\s)(?P<value>.*)$', re.M)
DESCRIPTION_RE = re.compile(r'(?:description\s)(?P<value>.*)$', re.M)


class Vrfs(EntityCollection):
    """The Vrfs class provides a configuration resource for VRFs

    The Vrfs class is derived from ResourceBase a standard set of methods
    for working with VRF configurations on an EOS node.

    """

    def get(self, value):
        """Returns the VRF configuration as a resource dict.

        Args:
            value (string): The vrf name to retrieve from the
                running configuration.

        Returns:
            A Python dict object containing the VRF attributes as
                key/value pairs.

        """
        if self.version_number >= '4.23':
            config = self.get_block('vrf instance %s' % value)
        else:
            config = self.get_block('vrf definition %s' % value)
        if not config:
            return None
        response = dict(vrf_name=value)
        response.update(self._parse_rd(config))
        response.update(self._parse_description(config))
        config = self.get_block('no ip routing vrf %s' % value)
        if config:
            response['ipv4_routing'] = False
        else:
            response['ipv4_routing'] = True
        config = self.get_block('no ipv6 unicast-routing vrf %s' % value)
        if config:
            response['ipv6_routing'] = False
        else:
            response['ipv6_routing'] = True

        return response

    def _parse_rd(self, config):
        """ _parse_rd scans the provided configuration block and extracts
        the vrf rd. The return dict is intended to be merged into the response
        dict.

        Args:
            config (str): The vrf configuration block from the nodes running
                configuration

        Returns:
            dict: resource dict attribute
        """
        match = RD_RE.search(config)
        if match:
            value = match.group('value')
        else:
            value = match
        return dict(rd=value)

    def _parse_description(self, config):
        """ _parse_description scans the provided configuration block and
        extracts the vrf description value. The return dict is intended to
        be merged into the response dict.

        Args:
            config (str): The vrf configuration block from the nodes
                running configuration

        Returns:
            dict: resource dict attribute
        """
        value = DESCRIPTION_RE.search(config).group('value')
        return dict(description=value)

    def getall(self):
        """Returns a dict object of all VRFs in the running-config

        Returns:
            A dict object of VRF attributes

        """
        if self.version_number >= '4.23':
            vrfs_re = re.compile(r'(?<=^vrf instance\s)(\w+)', re.M)
        else:
            vrfs_re = re.compile(r'(?<=^vrf definition\s)(\w+)', re.M)

        response = dict()
        for vrf in vrfs_re.findall(self.config):
            response[vrf] = self.get(vrf)
        return response

    def create(self, vrf_name, rd=None):
        """ Creates a new VRF resource

        Note: A valid RD has the following format admin_ID:local_assignment.
            The admin_ID can be an AS number or globally assigned IPv4 address.
            The local_assignment can be an integer between 0-65,535 if the
            admin_ID is an IPv4 address and can be between 0-4,294,967,295 if
            the admin_ID is an AS number. If the admin_ID is an AS number the
            local_assignment could also be in the form of an IPv4 address.

        Args:
            vrf_name (str): The VRF name to create
            rd (str): The value to configure the vrf rd

        Returns:
            True if create was successful otherwise False
        """
        if self.version_number >= '4.23':
            commands = ['vrf instance %s' % vrf_name]
        else:
            commands = ['vrf definition %s' % vrf_name]
        if rd:
            commands.append('rd %s' % rd)
        return self.configure(commands)

    def delete(self, vrf_name):
        """ Deletes a VRF from the running configuration

        Args:
            vrf_name (str): The VRF name to delete

        Returns:
            True if the operation was successful otherwise False
        """
        if self.version_number >= '4.23':
            command = 'no vrf instance %s' % vrf_name
        else:
            command = 'no vrf definition %s' % vrf_name
        return self.configure(command)

    def default(self, vrf_name):
        """ Defaults the VRF configuration for given name

        Args:
            vrf_name (str): The VRF name to default

        Returns:
            True if the operation was successful otherwise False
        """
        if self.version_number >= '4.23':
            command = 'default vrf instance %s' % vrf_name
        else:
            command = 'default vrf definition %s' % vrf_name
        return self.configure(command)

    def configure_vrf(self, vrf_name, commands):
        """ Configures the specified VRF using commands

        Args:
            vrf_name (str): The VRF name to configure
            commands: The list of commands to configure

        Returns:
            True if the commands completed successfully
        """
        commands = make_iterable(commands)
        if self.version_number >= '4.23':
            commands.insert(0, 'vrf instance %s' % vrf_name)
        else:
            commands.insert(0, 'vrf definition %s' % vrf_name)

        return self.configure(commands)

    def set_rd(self, vrf_name, rd):
        """ Configures the VRF rd (route distinguisher)

        Note: A valid RD has the following format admin_ID:local_assignment.
            The admin_ID can be an AS number or globally assigned IPv4 address.
            The local_assignment can be an integer between 0-65,535 if the
            admin_ID is an IPv4 address and can be between 0-4,294,967,295 if
            the admin_ID is an AS number. If the admin_ID is an AS number the
            local_assignment could also be in the form of an IPv4 address.

        Args:
            vrf_name (str): The VRF name to set rd for
            rd (str): The value to configure the vrf rd

        Returns:
            True if the operation was successful otherwise False
        """
        cmds = self.command_builder('rd', value=rd)
        return self.configure_vrf(vrf_name, cmds)

    def set_description(self, vrf_name, description=None, default=False,
                        disable=False):
        """ Configures the VRF description

        Args:
            vrf_name (str): The VRF name to configure
            description(str): The string to set the vrf description to
            default (bool): Configures the vrf description to its default value
            disable (bool): Negates the vrf description

        Returns:
            True if the operation was successful otherwise False
        """
        cmds = self.command_builder('description', value=description,
                                    default=default, disable=disable)
        return self.configure_vrf(vrf_name, cmds)

    def set_ipv4_routing(self, vrf_name, default=False, disable=False):
        """ Configures ipv4 routing for the vrf

        Args:
            vrf_name (str): The VRF name to configure
            default (bool): Configures ipv4 routing for the vrf value to
                default if this value is true
            disable (bool): Negates the ipv4 routing for the vrf if set to true

        Returns:
            True if the operation was successful otherwise False

        """
        cmd = 'ip routing vrf %s' % vrf_name
        if default:
            cmd = 'default %s' % cmd
        elif disable:
            cmd = 'no %s' % cmd
        cmd = make_iterable(cmd)
        return self.configure(cmd)

    def set_ipv6_routing(self, vrf_name, default=False, disable=False):
        """ Configures ipv6 unicast routing for the vrf

        Args:
            vrf_name (str): The VRF name to configure
            default (bool): Configures ipv6 unicast routing for the vrf value
                to default if this value is true
            disable (bool): Negates the ipv6 unicast routing for the vrf if set
                to true

        Returns:
            True if the operation was successful otherwise False

        """
        cmd = 'ipv6 unicast-routing vrf %s' % vrf_name
        if default:
            cmd = 'default %s' % cmd
        elif disable:
            cmd = 'no %s' % cmd
        cmd = make_iterable(cmd)
        return self.configure(cmd)

    def set_interface(self, vrf_name, interface, default=False, disable=False):
        """ Adds a VRF to an interface

        Notes:
            Requires interface to be in routed mode. Must apply ip address
            after VRF has been applied. This feature can also be accessed
            through the interfaces api.

        Args:
            vrf_name (str): The VRF name to configure
            interface (str): The interface to add the VRF too
            default (bool): Set interface VRF forwarding to default
            disable (bool): Negate interface VRF forwarding

        Returns:
            True if the operation was successful otherwise False
        """
        cmds = ['interface %s' % interface]
        if self.version_number >= '4.23':
            cmds.append(self.command_builder('vrf', value=vrf_name,
                        default=default, disable=disable))
        else:
            cmds.append(self.command_builder('vrf forwarding', value=vrf_name,
                        default=default, disable=disable))
        return self.configure(cmds)


def instance(node):
    """Returns an instance of Vrfs

    This method will create and return an instance of the Vrfs object passing
    the value of API to the object.  The instance method is required for the
    resource to be autoloaded by the Node object

    Args:
        node (Node): The node argument passes an instance of Node to the
            resource
    """
    return Vrfs(node)



#
# Copyright (c) 2016, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Module for working with OSPF configuration in EOS

This module provides an API for creating/modifying/deleting
OSPF configurations

"""

import re
from pyeapi.api import Entity
from pyeapi.utils import make_iterable

class Ospf(Entity):
    """ The Ospf class implements global Ospf router configuration
    """

    def __init__(self, *args, **kwargs):
        super(Ospf, self).__init__(*args, **kwargs)
        pass

    def get(self, vrf=None):
        """Returns the OSPF routing configuration

           Args:
                vrf (str): VRF name to return OSPF routing config for
           Returns:
               dict:
                    keys:
                         router_id (int): OSPF router-id

                         vrf (str): VRF of the OSPF process
                         networks (dict): All networks that
                         are advertised in OSPF

                         ospf_process_id (int): OSPF proc id

                         redistribution (dict): All protocols that
                         are configured to be redistributed in OSPF

                         shutdown (bool): Gives the current shutdown
                         off the process
        """
        match = '^router ospf .*'
        if vrf:
            match += ' vrf %s' % vrf
        config = self.get_block(match)
        if not config:
            return None

        response = dict()
        response.update(self._parse_router_id(config))
        response.update(self._parse_vrf(config))
        response.update(self._parse_networks(config))
        response.update(self._parse_ospf_process_id(config))
        response.update(self._parse_redistribution(config))
        response.update(self._parse_shutdown(config))

        return response

    def _parse_ospf_process_id(self, config):
        """Parses config file for the OSPF proc ID

           Args:
               config(str):  Running configuration
           Returns:
               dict: key: ospf_process_id (int)
        """
        match = re.search(r'^router ospf (\d+)', config)
        return dict(ospf_process_id=int(match.group(1)))

    def _parse_vrf(self, config):
        """Parses config file for the OSPF vrf name

           Args:
               config(str):  Running configuration
           Returns:
               dict: key: ospf_vrf (str)
        """
        match = re.search(r'^router ospf \d+ vrf (\w+)', config)
        if match:
            return dict(vrf=match.group(1))
        return dict(vrf='default')

    def _parse_router_id(self, config):
        """Parses config file for the OSPF router ID

           Args:
               config(str):  Running configuration
           Returns:
               dict: key: router_id (str)
        """
        match = re.search(r'router-id ([^\s]+)', config)
        value = match.group(1) if match else None
        return dict(router_id=value)

    def _parse_networks(self, config):
        """Parses config file for the networks advertised
           by the OSPF process

           Args:
               config(str):  Running configuration
           Returns:
               list: dict:
                         keys: network (str)
                               netmask (str)
                               area (str)
        """

        networks = list()
        regexp = r'network (.+)/(\d+) area (\d+\.\d+\.\d+\.\d+)'
        matches = re.findall(regexp, config)
        for (network, netmask, area) in matches:
            networks.append(dict(network=network, netmask=netmask, area=area))
        return dict(networks=networks)

    def _parse_redistribution(self, config):
        """Parses config file for the OSPF router ID

           Args:
               config (str):  Running configuration
           Returns:
               list: dict:
                         keys: protocol (str)
                               route-map (optional) (str)
        """
        redistributions = list()
        regexp = r'redistribute .*'
        matches = re.findall(regexp, config)
        for line in matches:
            ospf_redist = line.split()
            if len(ospf_redist) == 2:
                # simple redist: eg 'redistribute bgp'
                protocol = ospf_redist[1]
                redistributions.append(dict(protocol=protocol))
            if len(ospf_redist) == 4:
                # complex redist eg 'redistribute bgp route-map NYSE-RP-MAP'
                protocol = ospf_redist[1]
                route_map_name = ospf_redist[3]
                redistributions.append(dict(protocol=protocol,
                                       route_map=route_map_name))
        return dict(redistributions=redistributions)

    def _parse_shutdown(self, config):
        """Parses config file for the OSPF router ID

           Args:
               config(str):  Running configuration
           Returns:
               dict: key: shutdown (bool)
        """

        value = 'no shutdown' in config
        return dict(shutdown=not value)

    def set_shutdown(self):
        """Shutdowns the OSPF process

           Args:
               None
           Returns:
              bool: True if the commands are completed successfully
        """

        cmd = 'shutdown'
        return self.configure_ospf(cmd)

    def set_no_shutdown(self):
        """Removes the shutdown property from the OSPF process

           Args:
               None
           Returns:
              bool: True if the commands are completed successfully
        """


        cmd = 'no shutdown'
        return self.configure_ospf(cmd)

    def delete(self):
        """Removes the entire ospf process from the running configuration

           Args:
               None
           Returns:
               bool: True if the command completed succssfully
        """
        config = self.get()
        if not config:
            return True
        command = 'no router ospf {}'.format(config['ospf_process_id'])
        return self.configure(command)

    def create(self, ospf_process_id, vrf=None):
        """Creates a OSPF process in the specified VRF or the default VRF.

           Args:
                ospf_process_id (str): The OSPF process Id value
                vrf (str): The VRF to apply this OSPF process to
           Returns:
                bool: True if the command completed successfully
           Exception:
                ValueError: If the ospf_process_id passed in less
                            than 0 or greater than 65536
        """
        value = int(ospf_process_id)
        if not 0 < value < 65536:
            raise ValueError('ospf as must be between 1 and 65535')
        command = 'router ospf {}'.format(ospf_process_id)
        if vrf:
            command += ' vrf %s' % vrf
        return self.configure(command)

    def configure_ospf(self, cmd):
        """Allows for a list of OSPF subcommands to be configured"

           Args:
               cmd: (list or str): Subcommand to be entered
           Returns:
               bool: True if all the commands completed successfully
        """
        config = self.get()
        cmds = ['router ospf {}'.format(config['ospf_process_id'])]
        cmds.extend(make_iterable(cmd))
        return super(Ospf, self).configure(cmds)

    def set_router_id(self, value=None, default=False, disable=False):
        """Controls the router id property for the OSPF Proccess

           Args:
               value (str): The router-id value
               default (bool): Controls the use of the default keyword
               disable (bool): Controls the use of the no keyword
           Returns:
               bool: True if the commands are completed successfully
        """
        cmd = self.command_builder('router-id', value=value,
                                   default=default, disable=disable)
        return self.configure_ospf(cmd)

    def add_network(self, network, netmask, area=0):
        """Adds a network to be advertised by OSPF

           Args:
               network (str):  The network to be advertised in dotted decimal
                               notation
               netmask (str):  The netmask to configure
               area (str):  The area the network belongs to.
                            By default this value is 0
           Returns:
               bool: True if the command completes successfully
           Exception:
               ValueError: This will get raised if network or netmask
                           are not passed to the method
        """
        if network == '' or netmask == '':
            raise ValueError('network and mask values '
                             'may not be empty')
        cmd = 'network {}/{} area {}'.format(network, netmask, area)
        return self.configure_ospf(cmd)

    def remove_network(self, network, netmask, area=0):
        """Removes a network advertisment by OSPF

           Args:
               network (str):  The network to be removed in dotted decimal
                               notation
               netmask (str):  The netmask to configure
               area (str):  The area the network belongs to.
                            By default this value is 0
           Returns:
               bool: True if the command completes successfully
           Exception:
               ValueError: This will get raised if network or netmask
                           are not passed to the method
        """

        if network == '' or netmask == '':
            raise ValueError('network and mask values '
                             'may not be empty')
        cmd = 'no network {}/{} area {}'.format(network, netmask, area)
        return self.configure_ospf(cmd)

    def add_redistribution(self, protocol, route_map_name=None):
        """Adds a protocol redistribution to OSPF

           Args:
               protocol (str):  protocol to redistribute
               route_map_name (str): route-map to be used to
                                     filter the protocols
           Returns:
               bool: True if the command completes successfully
           Exception:
               ValueError:  This will be raised if the protocol pass is not one
                            of the following: [rip, bgp, static, connected]
        """
        protocols = ['bgp', 'rip', 'static', 'connected']
        if protocol not in protocols:
            raise ValueError('redistributed protocol must be'
                             'bgp, connected, rip or static')
        if route_map_name is None:
            cmd = 'redistribute {}'.format(protocol)
        else:
            cmd = 'redistribute {} route-map {}'.format(protocol,
                                                        route_map_name)
        return self.configure_ospf(cmd)

    def remove_redistribution(self, protocol):
        """Removes a protocol redistribution to OSPF

           Args:
               protocol (str):  protocol to redistribute
               route_map_name (str): route-map to be used to
                                     filter the protocols
           Returns:
               bool: True if the command completes successfully
           Exception:
               ValueError:  This will be raised if the protocol pass is not one
                            of the following: [rip, bgp, static, connected]
        """

        protocols = ['bgp', 'rip', 'static', 'connected']
        if protocol not in protocols:
            raise ValueError('redistributed protocol must be'
                             'bgp, connected, rip or static')
        cmd = 'no redistribute {}'.format(protocol)
        return self.configure_ospf(cmd)

def instance(api):
    """Returns an instance of Ospf
    """
    return Ospf(api)



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Module for working with EOS static routes

The staticroute resource provides configuration management of static
route resources on an EOS node. It provides the following class
implementations:

    * StaticRoute - Configure static routes in EOS

StaticRoute Attributes:
    ip_dest (string): The ip address of the destination in the
        form of A.B.C.D/E

    next_hop (string): The next hop interface or ip address
    next_hop_ip (string): The next hop address on destination interface
    distance (int): Administrative distance for this route
    tag (int): Route tag
    route_name (string): Route name

Notes:
    The 'default' prefix function of the 'ip route' command,
    'default ip route ...', currently equivalent to the 'no ip route ...'
    command.
"""

import re

from pyeapi.api import EntityCollection

# Define the regex to match ip route lines (by lines in regex):
#   'ip route' header
#   ip_dest
#   next_hop
#   next_hop_ip
#   distance
#   tag
#   name
ROUTES_RE = re.compile(r'(?<=^ip route)'
                       r' (\d+\.\d+\.\d+\.\d+\/\d+)'
                       r' (\d+\.\d+\.\d+\.\d+|\S+)'
                       r'(?: (\d+\.\d+\.\d+\.\d+))?'
                       r' (\d+)'
                       r'(?: tag (\d+))?'
                       r'(?: name (\S+))?', re.M)


class StaticRoute(EntityCollection):
    """The StaticRoute class provides a configuration instance
    for working with static routes

    """

    def __str__(self):
        return 'StaticRoute'

    def get(self, name):
        """Retrieves the ip route information for the destination
        ip address specified.

        Args:
            name (string): The ip address of the destination in the
                form of A.B.C.D/E

        Returns:
            dict: An dict object of static route entries in the form::

                { ip_dest:
                    { next_hop:
                        { next_hop_ip:
                            { distance:
                                { 'tag': tag,
                                  'route_name': route_name
                                }
                            }
                        }
                    }
                }

            If the ip address specified does not have any associated
            static routes, then None is returned.

        Notes:
            The keys ip_dest, next_hop, next_hop_ip, and distance in
            the returned dictionary are the values of those components
            of the ip route specification. If a route does not contain
            a next_hop_ip, then that key value will be set as 'None'.
        """

        # Return the route configurations for the specified ip address,
        # or None if its not found
        return self.getall().get(name)

    def getall(self):
        """Return all ip routes configured on the switch as a resource dict

        Returns:
            dict: An dict object of static route entries in the form::

                { ip_dest:
                    { next_hop:
                        { next_hop_ip:
                            { distance:
                                { 'tag': tag,
                                  'route_name': route_name
                                }
                            }
                        }
                    }
                }

            If the ip address specified does not have any associated
            static routes, then None is returned.

        Notes:
            The keys ip_dest, next_hop, next_hop_ip, and distance in
            the returned dictionary are the values of those components
            of the ip route specification. If a route does not contain
            a next_hop_ip, then that key value will be set as 'None'.
        """

        # Find all the ip routes in the config
        matches = ROUTES_RE.findall(self.config)

        # Parse the routes and add them to the routes dict
        routes = dict()
        for match in matches:

            # Get the four identifying components
            ip_dest = match[0]
            next_hop = match[1]
            next_hop_ip = None if match[2] == '' else match[2]
            distance = int(match[3])

            # Create the data dict with the remaining components
            data = {}
            data['tag'] = None if match[4] == '' else int(match[4])
            data['route_name'] = None if match[5] == '' else match[5]

            # Build the complete dict entry from the four components
            # and the data.
            # temp_dict = parent_dict[key] = parent_dict.get(key, {})
            # This creates the keyed dict in the parent_dict if it doesn't
            # exist, or reuses the existing keyed dict.
            # The temp_dict is used to make things more readable.
            ip_dict = routes[ip_dest] = routes.get(ip_dest, {})
            nh_dict = ip_dict[next_hop] = ip_dict.get(next_hop, {})
            nhip_dict = nh_dict[next_hop_ip] = nh_dict.get(next_hop_ip, {})
            nhip_dict[distance] = data

        return routes

    def create(self, ip_dest, next_hop, **kwargs):
        """Create a static route

        Args:
            ip_dest (string): The ip address of the destination in the
                form of A.B.C.D/E
            next_hop (string): The next hop interface or ip address
            **kwargs['next_hop_ip'] (string): The next hop address on
                destination interface
            **kwargs['distance'] (string): Administrative distance for this
                route
            **kwargs['tag'] (string): Route tag
            **kwargs['route_name'] (string): Route name

        Returns:
            True if the operation succeeds, otherwise False.
        """

        # Call _set_route with delete and default set to False
        return self._set_route(ip_dest, next_hop, **kwargs)

    def delete(self, ip_dest, next_hop, **kwargs):
        """Delete a static route

        Args:
            ip_dest (string): The ip address of the destination in the
                form of A.B.C.D/E
            next_hop (string): The next hop interface or ip address
            **kwargs['next_hop_ip'] (string): The next hop address on
                destination interface
            **kwargs['distance'] (string): Administrative distance for this
                route
            **kwargs['tag'] (string): Route tag
            **kwargs['route_name'] (string): Route name

        Returns:
            True if the operation succeeds, otherwise False.
        """

        # Call _set_route with the delete flag set to True
        kwargs.update({'delete': True})
        return self._set_route(ip_dest, next_hop, **kwargs)

    def default(self, ip_dest, next_hop, **kwargs):
        """Set a static route to default (i.e. delete the matching route)

        Args:
            ip_dest (string): The ip address of the destination in the
                form of A.B.C.D/E
            next_hop (string): The next hop interface or ip address
            **kwargs['next_hop_ip'] (string): The next hop address on
                destination interface
            **kwargs['distance'] (string): Administrative distance for this
                route
            **kwargs['tag'] (string): Route tag
            **kwargs['route_name'] (string): Route name

        Returns:
            True if the operation succeeds, otherwise False.
        """

        # Call _set_route with the default flag set to True
        kwargs.update({'default': True})
        return self._set_route(ip_dest, next_hop, **kwargs)

    def set_tag(self, ip_dest, next_hop, **kwargs):
        """Set the tag value for the specified route

        Args:
            ip_dest (string): The ip address of the destination in the
                form of A.B.C.D/E
            next_hop (string): The next hop interface or ip address
            **kwargs['next_hop_ip'] (string): The next hop address on
                destination interface
            **kwargs['distance'] (string): Administrative distance for this
                route
            **kwargs['tag'] (string): Route tag
            **kwargs['route_name'] (string): Route name

        Returns:
            True if the operation succeeds, otherwise False.

        Notes:
            Any existing route_name value must be included in call to
                set_tag, otherwise the tag will be reset
                by the call to EOS.
        """

        # Call _set_route with the new tag information
        return self._set_route(ip_dest, next_hop, **kwargs)

    def set_route_name(self, ip_dest, next_hop, **kwargs):
        """Set the route_name value for the specified route

        Args:
            ip_dest (string): The ip address of the destination in the
                form of A.B.C.D/E
            next_hop (string): The next hop interface or ip address
            **kwargs['next_hop_ip'] (string): The next hop address on
                destination interface
            **kwargs['distance'] (string): Administrative distance for this
                route
            **kwargs['tag'] (string): Route tag
            **kwargs['route_name'] (string): Route name

        Returns:
            True if the operation succeeds, otherwise False.

        Notes:
            Any existing tag value must be included in call to
                set_route_name, otherwise the tag will be reset
                by the call to EOS.
        """

        # Call _set_route with the new route_name information
        return self._set_route(ip_dest, next_hop, **kwargs)

    def _build_commands(self, ip_dest, next_hop, **kwargs):
        """Build the EOS command string for ip route interactions.

        Args:
            ip_dest (string): The ip address of the destination in the
                form of A.B.C.D/E
            next_hop (string): The next hop interface or ip address
            **kwargs['next_hop_ip'] (string): The next hop address on
                destination interface
            **kwargs['distance'] (string): Administrative distance for this
                route
            **kwargs['tag'] (string): Route tag
            **kwargs['route_name'] (string): Route name

        Returns the ip route command string to be sent to the switch for
        the given set of parameters.
        """

        commands = "ip route %s %s" % (ip_dest, next_hop)

        next_hop_ip = kwargs.get('next_hop_ip', None)
        distance = kwargs.get('distance', None)
        tag = kwargs.get('tag', None)
        route_name = kwargs.get('route_name', None)

        if next_hop_ip is not None:
            commands += " %s" % next_hop_ip
        if distance is not None:
            commands += " %s" % distance
        if tag is not None:
            commands += " tag %s" % tag
        if route_name is not None:
            commands += " name %s" % route_name

        return commands

    def _set_route(self, ip_dest, next_hop, **kwargs):
        """Configure a static route

        Args:
            ip_dest (string): The ip address of the destination in the
                form of A.B.C.D/E
            next_hop (string): The next hop interface or ip address
            **kwargs['next_hop_ip'] (string): The next hop address on
                destination interface
            **kwargs['distance'] (string): Administrative distance for this
                route
            **kwargs['tag'] (string): Route tag
            **kwargs['route_name'] (string): Route name
            **kwargs['delete'] (boolean): If true, deletes the specified route
                instead of creating or setting values for the route
            **kwargs['default'] (boolean): If true, defaults the specified
                route instead of creating or setting values for the route

        Returns:
            True if the operation succeeds, otherwise False.
        """

        commands = self._build_commands(ip_dest, next_hop, **kwargs)

        delete = kwargs.get('delete', False)
        default = kwargs.get('default', False)

        # Prefix with 'no' if delete is set
        if delete:
            commands = "no " + commands
        # Or with 'default' if default is setting
        else:
            if default:
                commands = "default " + commands

        return self.configure(commands)


def instance(node):
    """Returns an instance of StaticRoute

    This method will create and return an instance of the StaticRoute
    object passing the value of API to the object. The instance method
    is required for the resource to be autoloaded by the Node object

    Args:
        node (Node): The node argument passes an instance of Node to the
            resource
    """
    return StaticRoute(node)



#
# Copyright (c) 2015, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Module for managing the VARP configuration in EOS

This module provides an API for configuring VARP resources using
EOS and eAPI.

Arguments:
    name (string): The interface name the configuration is in reference
        to.  The interface name is the full interface identifier

    address (string): The interface IP address in the form of
        address/len.

    mtu (integer): The interface MTU value.  The MTU value accepts
        integers in the range of 68 to 65535 bytes
"""

import re

from pyeapi.api import EntityCollection


class Varp(EntityCollection):

    def __init__(self, *args, **kwargs):
        super(Varp, self).__init__(*args, **kwargs)
        self._interfaces = None

    @property
    def interfaces(self):
        if self._interfaces is not None:
            return self._interfaces
        self._interfaces = VarpInterfaces(self.node)
        return self._interfaces

    def get(self):
        """Returns the current VARP configuration

        The Varp resource returns the following:

            * mac_address (str): The virtual-router mac address
            * interfaces (dict): A list of the interfaces that have a
                                 virtual-router address configured.

        Return:
            A Python dictionary object of key/value pairs that represents
            the current configuration of the node.  If the specified
            interface does not exist then None is returned::

                {
                    "mac_address": "aa:bb:cc:dd:ee:ff",
                    "interfaces": {
                        "Vlan100": {
                            "addresses": [ "1.1.1.1", "2.2.2.2"]
                        },
                        "Vlan200": [...]
                    }
                }
        """
        resource = dict()
        resource.update(self._parse_mac_address())
        resource.update(self._parse_interfaces())
        return resource

    def _parse_mac_address(self):
        mac_address_re = re.compile(r'^ip\svirtual-router\smac-address\s'
                                    r'((?:[a-f0-9]{2}:){5}[a-f0-9]{2})$', re.M)
        mac = mac_address_re.search(self.config)
        mac = mac.group(1) if mac else None
        return dict(mac_address=mac)

    def _parse_interfaces(self):
        interfaces = VarpInterfaces(self.node).getall()
        return dict(interfaces=interfaces)

    def set_mac_address(self, mac_address=None, default=False, disable=False):
        """ Sets the virtual-router mac address

        This method will set the switch virtual-router mac address. If a
        virtual-router mac address already exists it will be overwritten.

        Args:
            mac_address (string): The mac address that will be assigned as
                the virtual-router mac address. This should be in the format,
                aa:bb:cc:dd:ee:ff.
            default (bool): Sets the virtual-router mac address to the system
                default (which is to remove the configuration line).
            disable (bool): Negates the virtual-router mac address using
                the system no configuration command

        Returns:
            True if the set operation succeeds otherwise False.
        """
        base_command = 'ip virtual-router mac-address'
        if not default and not disable:
            if mac_address is not None:
                # Check to see if mac_address matches expected format
                if not re.match(r'(?:[a-f0-9]{2}:){5}[a-f0-9]{2}',
                                mac_address):
                    raise ValueError('mac_address must be formatted like:'
                                     'aa:bb:cc:dd:ee:ff')
            else:
                raise ValueError('mac_address must be a properly formatted '
                                 'address string')
        if default or disable and not mac_address:
            current_mac = self._parse_mac_address()
            if current_mac['mac_address']:
                base_command = base_command + ' ' + current_mac['mac_address']
        commands = self.command_builder(base_command, value=mac_address,
                                        default=default, disable=disable)
        return self.configure(commands)


class VarpInterfaces(EntityCollection):
    """The VarpInterfaces class helps manage interfaces with
    virtual-router configuration.
    """
    def get(self, name):
        interface_re = r'interface\s%s' % name
        config = self.get_block(interface_re)

        if not config:
            return None

        resource = dict(addresses=dict())
        resource.update(self._parse_virtual_addresses(config))
        return resource

    def getall(self):
        resources = dict()
        interfaces_re = re.compile(r'^interface\s(Vlan\d+)$', re.M)
        for name in interfaces_re.findall(self.config):
            interface_detail = self.get(name)
            if interface_detail:
                resources[name] = interface_detail
        return resources

    def set_addresses(self, name, addresses=None, default=False,
                      disable=False):

        commands = list()
        commands.append('interface %s' % name)

        if default:
            commands.append('default ip virtual-router address')
        elif disable:
            commands.append('no ip virtual-router address')
        elif addresses is not None:
            try:
                current_addresses = self.get(name)['addresses']
            except Exception:
                current_addresses = []

            # remove virtual-router addresses not present in addresses list
            for entry in set(current_addresses).difference(addresses):
                commands.append('no ip virtual-router address %s' % entry)

            # add new set virtual-router addresses that werent present
            for entry in set(addresses).difference(current_addresses):
                commands.append('ip virtual-router address %s' % entry)
        else:
            commands.append('no ip virtual-router address')

        return self.configure(commands) if commands else True

    def _parse_virtual_addresses(self, config):
        virt_ip_re = re.compile(r'^\s+ip\svirtual-router\saddress\s(\S+)$',
                                re.M)
        return dict(addresses=virt_ip_re.findall(config))


def instance(node):
    """Returns an instance of Ipinterfaces

    This method will create and return an instance of the Varp object
    passing the value of node to the instance.  This function is required
    for the resource to be autoloaded by the Node object

    Args:
        node (Node): The node argument provides an instance of Node to
            the Varp instance
    """
    return Varp(node)



#
# Copyright (c) 2015, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Module for working with the global system in EOS

This module provides an API for working with the global system settings
in EOS.   It provides the following class implementations:

    * System -- Configures global system settings

System Attributes:
    hostname (string): The hostname of the node as configured in the
        running-configuration.

"""

import re

from pyeapi.api import Entity


class System(Entity):
    """The System class implements global config for the node

    Global configuration settings include those thaat identify the node
    and provide node level configuration such as hostname
    """

    def get(self):
        """Returns the system configuration abstraction

        The System resource returns the following:

            * hostname (str): The hostname value

        Returns:
            dict: Represents the node's system configuration
        """
        resource = dict()
        resource.update(self._parse_hostname())
        resource.update(self._parse_iprouting())
        resource.update(self._parse_banners())

        return resource

    def _parse_hostname(self):
        """Parses the global config and returns the hostname value

        Returns:
            dict: The configured value for hostname.  The returned dict
                object is intended to be merged into the resource dict
        """
        value = 'localhost'
        match = re.search(r'^hostname ([^\s]+)$', self.config, re.M)
        if match:
            value = match.group(1)
        return dict(hostname=value)

    def _parse_iprouting(self):
        """Parses the global config and returns the ip routing value

        Returns:
            dict: The configure value for ip routing.  The returned dict
                object is intendd to be merged into the resource dict
        """
        value = 'no ip routing' not in self.config
        return dict(iprouting=value)

    def _parse_banners(self):
        """Parses the global config and returns the value for both motd
            and login banners.

        Returns:
           dict: The configure value for modtd and login banners. If the
                  banner is not set it will return a value of None for that
                  key. The returned dict object is intendd to be merged
                  into the resource dict
        """
        motd_value = login_value = None
        matches = re.findall(r'^banner\s+(login|motd)\s?$\n(.*?)$\nEOF$\n',
                             self.config, re.DOTALL | re.M)
        for match in matches:
            if match[0].strip() == "motd":
                motd_value = match[1]
            elif match[0].strip() == "login":
                login_value = match[1]

        return dict(banner_motd=motd_value, banner_login=login_value)

    def set_hostname(self, value=None, default=False, disable=False):
        """Configures the global system hostname setting

        EosVersion:
            4.13.7M

        Args:
            value (str): The hostname value
            default (bool): Controls use of the default keyword
            disable (bool): Controls the use of the no keyword

        Returns:
            bool: True if the commands are completed successfully
        """
        cmd = self.command_builder('hostname', value=value, default=default,
                                   disable=disable)
        return self.configure(cmd)

    def set_iprouting(self, value=None, default=False, disable=False):
        """Configures the state of global ip routing

        EosVersion:
            4.13.7M

        Args:
            value(bool): True if ip routing should be enabled or False if
                ip routing should be disabled
            default (bool): Controls the use of the default keyword
            disable (bool): Controls the use of the no keyword

        Returns:
            bool: True if the commands completed successfully otherwise False
        """
        if value is False:
            disable = True
        cmd = self.command_builder('ip routing', value=value, default=default,
                                   disable=disable)
        return self.configure(cmd)

    def set_banner(self, banner_type, value=None, default=False,
                   disable=False):
        """Configures system banners

        Args:
            banner_type(str): banner to be changed (likely login or motd)
            value(str): value to set for the banner
            default (bool): Controls the use of the default keyword
            disable (bool): Controls the use of the no keyword`

        Returns:
            bool: True if the commands completed successfully otherwise False
        """

        command_string = "banner %s" % banner_type
        if default is True or disable is True:
            cmd = self.command_builder(command_string, value=None,
                                       default=default, disable=disable)
            return self.configure(cmd)
        else:
            if not value.endswith("\n"):
                value = value + "\n"
            command_input = dict(cmd=command_string, input=value)
            return self.configure([command_input])


def instance(api):
    """Returns an instance of System
    """
    return System(api)



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Module for working with EOS access control list resources

This module provides an implementation for configuring and managing access
access control lists on Arista EOS nodes.  Access control lists can be
specified as either 'standard' or 'extended' ACLs.  This module provides the
following class implementations:

    * Acls -- The top-level class used to manage both standard and extended
        access control lists in EOS
    * StandardAcls -- Class that manages the set of standard ACLs
    * ExtendedAcls -- Class that manages the set of extended ACLs

"""
import re

import netaddr

from pyeapi.api import EntityCollection
from pyeapi.utils import ProxyCall

VALID_ACLS = frozenset(['standard', 'extended'])


def mask_to_prefixlen(mask):
    """Converts a subnet mask from dotted decimal to bit length

    Args:
        mask (str): The dotted decimal subnet mask to convert

    Returns:
        str: The subnet mask as a valid prefix length
    """
    mask = mask or '255.255.255.255'
    return netaddr.IPAddress(mask).netmask_bits()


def prefixlen_to_mask(prefixlen):
    """Converts a prefix length to a dotted decimal subnet mask

    Args:
        prefixlen (str): The prefix length value to convert

    Returns:
        str: The subt mask as a dotted decimal string
    """
    prefixlen = prefixlen or '32'
    addr = '0.0.0.0/%s' % prefixlen
    return str(netaddr.IPNetwork(addr).netmask)


class Acls(EntityCollection):

    def __init__(self, node, *args, **kwargs):
        super(Acls, self).__init__(node, *args, **kwargs)
        self._instances = dict()

    def get(self, name):
        return self.get_instance(name)[name]

    def getall(self):
        """Returns all ACLs in a dict object.

        Returns:
            A Python dictionary object containing all ACL
            configuration indexed by ACL name::

                {
                    "<ACL1 name>": {...},
                    "<ACL2 name>": {...}
                }

        """
        acl_re = re.compile(r'^ip access-list (?:(standard) )?(.+)$', re.M)
        response = {'standard': {}, 'extended': {}}
        for acl_type, name in acl_re.findall(self.config):
            acl = self.get(name)
            if acl_type and acl_type == 'standard':
                response['standard'][name] = acl
            else:
                response['extended'][name] = acl
        return response

    def __getattr__(self, name):
        return ProxyCall(self.marshall, name)

    def marshall(self, name, *args, **kwargs):
        acl_name = args[0]
        acl_instance = self.get_instance(acl_name)
        if not hasattr(acl_instance, name):
            raise AttributeError("'%s' object has no attribute '%s'" %
                                 (acl_instance, name))
        method = getattr(acl_instance, name)
        return method(*args, **kwargs)

    def get_instance(self, name):
        if name in self._instances:
            return self._instances[name]
        acl_re = re.compile(r'^ip access-list (?:(standard) )?(%s)$' % name,
                            re.M)
        match = acl_re.search(self.config)
        if match:
            acl_type = match.group(1) or 'extended'
            return self.create_instance(match.group(2), acl_type)
        return {name: None}

    def create_instance(self, name, acl_type):
        if acl_type not in VALID_ACLS:
            acl_type = 'standard'
        acl_instance = ACL_CLASS_MAP.get(acl_type)
        self._instances[name] = acl_instance(self.node)
        return self._instances[name]

    def create(self, name, type='standard'):
        # Create ACL instance for ACL type Standard or Extended then call
        # create method for specific ACL class.
        acl_instance = self.create_instance(name, type)
        return acl_instance.create(name)


class StandardAcls(EntityCollection):

    entry_re = re.compile(r'(\d+)'
                          r'(?: ([p|d]\w+))'
                          r'(?: (any))?'
                          r'(?: (host))?'
                          r'(?: ([0-9]+(?:\.[0-9]+){3}))?'
                          r'(?:/([0-9]{1,2}))?'
                          r'(?: ([0-9]+(?:\.[0-9]+){3}))?'
                          r'(?: (log))?')

    def get(self, name):
        config = self.get_block('ip access-list standard %s' % name)
        if not config:
            return None
        resource = dict(name=name, type='standard')
        resource.update(self._parse_entries(config))
        return resource

    def _parse_entries(self, config):
        entries = dict()
        for item in re.finditer(r'\d+ [p|d].*$', config, re.M):
            match = self.entry_re.match(item.group(0))
            (seq, act, anyip, host, ip, mlen, mask, log) = match.groups()
            entry = dict()
            entry['action'] = act
            entry['srcaddr'] = ip or '0.0.0.0'
            entry['srclen'] = mlen or mask_to_prefixlen(mask)
            entry['log'] = log is not None
            entries[seq] = entry
        return dict(entries=entries)

    def create(self, name):
        return self.configure('ip access-list standard %s' % name)

    def delete(self, name):
        return self.configure('no ip access-list standard %s' % name)

    def default(self, name):
        return self.configure('default ip access-list standard %s' % name)

    def update_entry(self, name, seqno, action, addr, prefixlen, log=False):
        cmds = ['ip access-list standard %s' % name]
        cmds.append('no %s' % seqno)
        entry = '%s %s %s/%s' % (seqno, action, addr, prefixlen)
        if log:
            entry += ' log'
        cmds.append(entry)
        cmds.append('exit')
        return self.configure(cmds)

    def add_entry(self, name, action, addr, prefixlen, log=False, seqno=None):
        cmds = ['ip access-list standard %s' % name]
        entry = '%s %s/%s' % (action, addr, prefixlen)
        if seqno is not None:
            entry = '%s %s' % (seqno, entry)
        if log:
            entry += ' log'
        cmds.append(entry)
        cmds.append('exit')
        return self.configure(cmds)

    def remove_entry(self, name, seqno):
        cmds = ['ip access-list standard %s' % name, 'no %s' % seqno, 'exit']
        return self.configure(cmds)


class ExtendedAcls(EntityCollection):

    entry_re = re.compile(r'(\d+)'
                          r'(?: ([p|d]\w+))'
                          r'(?: (\w+|\d+))'
                          r'(?: ([a|h]\w+))?'
                          r'(?: ([0-9]+(?:\.[0-9]+){3}))?'
                          r'(?:/([0-9]{1,2}))?'
                          r'(?: ((?:eq|gt|lt|neq|range) [\w-]+))?'
                          r'(?: ([a|h]\w+))?'
                          r'(?: ([0-9]+(?:\.[0-9]+){3}))?'
                          r'(?:/([0-9]{1,2}))?'
                          r'(?: ([0-9]+(?:\.[0-9]+){3}))?'
                          r'(?: ((?:eq|gt|lt|neq|range) [\w-]+))?'
                          r'(?: (.+))?')

    def get(self, name):
        config = self.get_block('ip access-list %s' % name)
        if not config:
            return None
        resource = dict(name=name, type='extended')
        resource.update(self._parse_entries(config))
        return resource

    def _parse_entries(self, config):
        entries = dict()
        for item in re.finditer(r'\d+ [p|d].*$', config, re.M):
            match = self.entry_re.match(item.group(0))
            if match:
                entry = dict()
                entry['action'] = match.group(2)
                entry['protocol'] = match.group(3)
                entry['srcaddr'] = match.group(5) or 'any'
                entry['srclen'] = match.group(6)
                entry['srcport'] = match.group(7)
                entry['dstaddr'] = match.group(9) or 'any'
                entry['dstlen'] = match.group(10)
                entry['dstport'] = match.group(12)
                entry['other'] = match.group(13)
                entries[match.group(1)] = entry
        return dict(entries=entries)

    def create(self, name):
        return self.configure('ip access-list %s' % name)

    def delete(self, name):
        return self.configure('no ip access-list %s' % name)

    def default(self, name):
        return self.configure('default ip access-list %s' % name)

    def update_entry(self, name, seqno, action, protocol, srcaddr,
                     srcprefixlen, dstaddr, dstprefixlen, log=False):
        cmds = ['ip access-list %s' % name]
        cmds.append('no %s' % seqno)
        entry = '%s %s %s %s/%s %s/%s' % (seqno, action, protocol, srcaddr,
                                          srcprefixlen, dstaddr, dstprefixlen)
        if log:
            entry += ' log'
        cmds.append(entry)
        cmds.append('exit')
        return self.configure(cmds)

    def add_entry(self, name, action, protocol, srcaddr, srcprefixlen,
                  dstaddr, dstprefixlen, log=False, seqno=None):
        cmds = ['ip access-list %s' % name]
        entry = '%s %s %s/%s %s/%s' % (action, protocol, srcaddr,
                                       srcprefixlen, dstaddr, dstprefixlen)
        if seqno is not None:
            entry = '%s %s' % (seqno, entry)
        if log:
            entry += ' log'
        cmds.append(entry)
        cmds.append('exit')
        return self.configure(cmds)

    def remove_entry(self, name, seqno):
        cmds = ['ip access-list %s' % name, 'no %s' % seqno, 'exit']
        return self.configure(cmds)


ACL_CLASS_MAP = {'standard': StandardAcls, 'extended': ExtendedAcls}


def instance(node):
    return Acls(node)



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Provides an abstract implementation for building API modules

This module provides a set of classes that are used to build API modules
that work with Node objects.  Using this module will allow the API
modules to be automatically loaded using the Node.api method.

The classes in this module should not be instantiated directly but rather
provide parent class for API implementations.  All API modules will
ultimately derive from BaseEntity which provides some common functions to
make building API modules easier.
"""

from collections.abc import Callable, Mapping
from pyeapi.eapilib import CommandError
from pyeapi.utils import make_iterable


class BaseEntity(object):
    """Base class for all resources to derive from

    This BaseEntity class should not be directly instatiated.  It is
    designed to be implemented by all resource classes to provide common
    methods.

    Attributes:
        node (Node): The node instance this resource will perform operations
            against for configuration
        config (Config): Returns an instance of Config with the nodes
            current running configuration
        error (CommandError): Holds the latest CommandError exception
            instance if raised

    Args:
        node (Node): An instance of Node
    """
    def __init__(self, node):
        self.node = node

    @property
    def version_number(self):
        return self.node.version_number

    @property
    def config(self):
        return self.node.running_config

    @property
    def error(self):
        return self.node.connection.error

    def get_block(self, parent, config='running_config'):
        """ Scans the config and returns a block of code

        Args:
            parent (str): The parent string to search the config for and
                return the block
            config (str): A text config string to be searched. Default
                is to search the running-config of the Node.

        Returns:
            A string object that represents the block from the config.  If
            the parent string is not found, then this method will
            return None.

        """
        try:
            parent = r'^%s$' % parent
            return self.node.section(parent, config=config)
        except TypeError:
            return None

    def configure(self, commands):
        """Sends the commands list to the node in config mode

        This method performs configuration the node using the array of
        commands specified.   This method wraps the configuration commands
        in a try/except block and stores any exceptions in the error
        property.

        Note:
            If the return from this method is False, use the error property
            to investigate the exception

        Args:
            commands (list): A list of commands to be sent to the node in
                config mode

        Returns:
            True if the commands are executed without exception otherwise
                False is returned
        """
        try:
            self.node.config(commands)
            return True
        except (CommandError):
            return False

    def command_builder(self, string, value=None, default=None, disable=None):
        """Builds a command with keywords

        Notes:
            Negating a command string by overriding 'value' with None or an
                assigned value that evalutates to false has been deprecated.
                Please use 'disable' to negate a command.

            Parameters are evaluated in the order 'default', 'disable', 'value'

        Args:
            string (str): The command string
            value (str): The configuration setting to subsititue into the
                command string. If value is a boolean and True, just the
                command string is used
            default (bool): Specifies the command should use the default
                keyword argument. Default preempts disable and value.
            disable (bool): Specifies the command should use the no
                keyword argument. Disable preempts value.

        Returns:
            A command string that can be used to configure the node
        """
        if default:
            return 'default %s' % string
        elif disable:
            return 'no %s' % string
        elif value is True:
            return string
        elif value:
            return '%s %s' % (string, value)
        else:
            return 'no %s' % string
            # -- above line to be deprecated and replaced with the error below
            # raise ValueError("abstract.command_builder: No value "
            #                  "received '%s'" % value)

    def configure_interface(self, name, commands):
        """Configures the specified interface with the commands

        Args:
            name (str): The interface name to configure
            commands: The commands to configure in the interface

        Returns:
            True if the commands completed successfully
        """
        commands = make_iterable(commands)
        commands.insert(0, 'interface %s' % name)
        return self.configure(commands)


class Entity(BaseEntity, Callable):
    """Abstract class for building Entity resources

    The Entity class provides an abstract implementation that allows for
    building an API configuration resource.  The Entity class should not
    be directly instantiated.  It is used in instances where a single config
    entity is appropriate in the configuration.

    Examples of Entity candidates include global spanning tree
    """
    def __call__(self):
        return self.get()

    def get(self):
        raise NotImplementedError


class EntityCollection(BaseEntity, Mapping):
    """Abstract class for building EntityCollection resources

    The EntityCollection class provides an abstract implementat that allows
    for building API configuration resources with multiple resources.  The
    EntityCollection class should not be directly instantiated.

    Examples of an EntityCollection candidate include VLANs and interfaces
    """

    def __call__(self):
        return self.getall()

    def __getitem__(self, value):
        return self.get(value)

    def __len__(self):
        return len(self.getall())

    def __iter__(self):
        return iter(self.getall())

    def getall(self):
        raise NotImplementedError

    def get(self, name, default=None):
        raise NotImplementedError



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Module for working with EOS VLAN resources

The Vlans resource provides configuration of VLAN resources for an EOS
node.

Parameters:

    name (string): The name parameter maps to the VLAN name in EOS.  Valid
        values include any consecutive sequence of numbers, letters and
        underscore up to the maximum number of characters.  This parameter
        is defaultable.
    state (string): The state parameter sets the operational state of
        the VLAN on the node.   It has two valid values: active or suspend.
        The state parameter is defaultable.
    trunk_groups (array): The trunk_groups parameter provides a list of
        trunk groups configured for this VLAN.  This parameter is
        defaultable.

"""

import re

from pyeapi.api import EntityCollection
from pyeapi.utils import make_iterable

VLAN_ID_RE = re.compile(r'(?:vlan\s)(?P<value>.*)$', re.M)
NAME_RE = re.compile(r'(?:name\s)(?P<value>.*)$', re.M)
STATE_RE = re.compile(r'(?:state\s)(?P<value>.*)$', re.M)
TRUNK_GROUP_RE = re.compile(r'(?:trunk\sgroup\s)(?P<value>.*)$', re.M)


def isvlan(value):
    """Checks if the argument is a valid VLAN

    A valid VLAN is an integer value in the range of 1 to 4094.  This
    function will test if the argument falls into the specified range and
    is considered a valid VLAN

    Args:
        value: The value to check if is a valid VLAN

    Returns:
        True if the supplied value is a valid VLAN otherwise False
    """
    try:
        value = int(value)
        return value in range(1, 4095)
    except ValueError:
        return False


class Vlans(EntityCollection):
    """The Vlans class provides a configuration resource for VLANs

    The Vlans class is derived from ResourceBase a standard set of methods
    for working with VLAN configurations on an EOS node.

    """

    def get(self, value):
        """Returns the VLAN configuration as a resource dict.

        Args:
            vid (string): The vlan identifier to retrieve from the
                running configuration.  Valid values are in the range
                of 1 to 4095

        Returns:
            A Python dict object containing the VLAN attributes as
                key/value pairs.

        """
        config = self.get_block('vlan %s' % value)
        if not config:
            return None

        response = dict(vlan_id=self._parse_vlan_id(config))
        response.update(self._parse_name(config))
        response.update(self._parse_state(config))
        response.update(self._parse_trunk_groups(config))

        return response

    def _parse_vlan_id(self, config):
        """ _parse_vlan_id scans the provided configuration block and extracts
        the vlan id.  The config block is expected to always return the
        vlan id.  The return dict is intended to be merged into the response
        dict.

        Args:
            config (str): The vlan configuration block from the nodes running
                configuration

        Returns:
            Str: vlan id (or range/list of vlan ids)
        """
        value = VLAN_ID_RE.search(config).group('value')
        return value

    def _parse_name(self, config):
        """ _parse_name scans the provided configuration block and extracts
        the vlan name.  The config block is expected to always return the
        vlan name.  The return dict is intended to be merged into the response
        dict.

        Args:
            config (str): The vlan configuration block from the nodes running
                configuration

        Returns:
            dict: resource dict attribute
        """
        value = NAME_RE.search(config).group('value')
        return dict(name=value)

    def _parse_state(self, config):
        """ _parse_state scans the provided configuration block and extracts
        the vlan state value.  The config block is expected to always return
        the vlan state config.  The return dict is inteded to be merged into
        the response dict.

        Args:
            config (str): The vlan configuration block from the nodes
                running configuration

        Returns:
            dict: resource dict attribute
        """
        value = STATE_RE.search(config).group('value')
        return dict(state=value)

    def _parse_trunk_groups(self, config):
        """ _parse_trunk_groups scans the provided configuration block and
        extracts all the vlan trunk groups.  If no trunk groups are configured
        an empty List is returned as the vlaue.  The return dict is intended
        to be merged into the response dict.

        Args:
            config (str): The vlan configuration block form the node's
                running configuration

        Returns:
            dict: resource dict attribute
        """
        values = TRUNK_GROUP_RE.findall(config)
        return dict(trunk_groups=values)

    def getall(self):
        """Returns a dict object of all Vlans in the running-config

        Returns:
            A dict object of Vlan attributes

        """
        # RE to find standalone and grouped (ranged, enumerated) vlans (#197)
        vlans_re = re.compile(r'(?<=^vlan\s)[\d,\-]+', re.M)

        response = dict()
        for vid in vlans_re.findall(self.config):
            response[vid] = self.get(vid)
        return response

    def create(self, vid):
        """ Creates a new VLAN resource

        Args:
            vid (str): The VLAN ID to create

        Returns:
            True if create was successful otherwise False
        """
        command = 'vlan %s' % vid
        return self.configure(command) if isvlan(vid) else False

    def delete(self, vid):
        """ Deletes a VLAN from the running configuration

        Args:
            vid (str): The VLAN ID to delete

        Returns:
            True if the operation was successful otherwise False
        """
        command = 'no vlan %s' % vid
        return self.configure(command) if isvlan(vid) else False

    def default(self, vid):
        """ Defaults the VLAN configuration

        .. code-block:: none

            default vlan <vlanid>

        Args:
            vid (str): The VLAN ID to default

        Returns:
            True if the operation was successful otherwise False
        """
        command = 'default vlan %s' % vid
        return self.configure(command) if isvlan(vid) else False

    def configure_vlan(self, vid, commands):
        """ Configures the specified Vlan using commands

        Args:
            vid (str): The VLAN ID to configure
            commands: The list of commands to configure

        Returns:
            True if the commands completed successfully
        """
        commands = make_iterable(commands)
        commands.insert(0, 'vlan %s' % vid)
        return self.configure(commands)

    def set_name(self, vid, name=None, default=False, disable=False):
        """ Configures the VLAN name

        EosVersion:
            4.13.7M

        Args:
            vid (str): The VLAN ID to Configures
            name (str): The value to configure the vlan name
            default (bool): Defaults the VLAN ID name
            disable (bool): Negates the VLAN ID name

        Returns:
            True if the operation was successful otherwise False
        """
        cmds = self.command_builder('name', value=name, default=default,
                                    disable=disable)
        return self.configure_vlan(vid, cmds)

    def set_state(self, vid, value=None, default=False, disable=False):
        """ Configures the VLAN state

        EosVersion:
            4.13.7M

        Args:
            vid (str): The VLAN ID to configure
            value (str): The value to set the vlan state to
            default (bool): Configures the vlan state to its default value
            disable (bool): Negates the vlan state

        Returns:
            True if the operation was successful otherwise False
        """
        cmds = self.command_builder('state', value=value, default=default,
                                    disable=disable)
        return self.configure_vlan(vid, cmds)

    def set_trunk_groups(self, vid, value=None, default=False, disable=False):
        """ Configures the list of trunk groups support on a vlan

        This method handles configuring the vlan trunk group value to default
        if the default flag is set to True.  If the default flag is set
        to False, then this method will calculate the set of trunk
        group names to be added and to be removed.

        EosVersion:
            4.13.7M

        Args:
            vid (str): The VLAN ID to configure
            value (str): The list of trunk groups that should be configured
                for this vlan id.
            default (bool): Configures the trunk group value to default if
                this value is true
            disable (bool): Negates the trunk group value if set to true

        Returns:
            True if the operation was successful otherwise False

        """
        if default:
            return self.configure_vlan(vid, 'default trunk group')
        if disable:
            return self.configure_vlan(vid, 'no trunk group')

        current_value = self.get(vid)['trunk_groups']
        failure = False

        value = make_iterable(value)

        for name in set(value).difference(current_value):
            if not self.add_trunk_group(vid, name):
                failure = True

        for name in set(current_value).difference(value):
            if not self.remove_trunk_group(vid, name):
                failure = True

        return not failure

    def add_trunk_group(self, vid, name):
        """ Adds a new trunk group to the Vlan in the running-config

        EosVersion:
            4.13.7M

        Args:
            vid (str): The VLAN ID to configure
            name (str): The trunk group to add to the list

        Returns:
            True if the operation was successful otherwise False
        """
        return self.configure_vlan(vid, 'trunk group %s' % name)

    def remove_trunk_group(self, vid, name):
        """ Removes a trunk group from the list of configured trunk groups
        for the specified VLAN ID

        EosVersion:
            4.13.7M

        Args:
            vid (str): The VLAN ID to configure
            name (str): The trunk group to add to the list

        Returns:
            True if the operation was successful otherwise False
        """
        return self.configure_vlan(vid, 'no trunk group %s' % name)


def instance(node):
    """Returns an instance of Vlans

    This method will create and return an instance of the Vlans object passing
    the value of API to the object.  The instance method is required for the
    resource to be autoloaded by the Node object

    Args:
        node (Node): The node argument passes an instance of Node to the
            resource
    """
    return Vlans(node)



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import os
import sys
import importlib
import inspect
import logging
import logging.handlers

from collections.abc import Iterable
from itertools import tee, zip_longest


_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.DEBUG)

# Create a handler to log messages to syslog
if sys.platform == "darwin":
    _syslog_handler = logging.handlers.SysLogHandler(address='/var/run/syslog')
else:
    _syslog_handler = logging.handlers.SysLogHandler()
_LOGGER.addHandler(_syslog_handler)

# Create a handler to log messages to stderr
_stderr_formatter = logging.Formatter('\n\n**** LOG NOTE ****\n%(message)s\n')
_stderr_handler = logging.StreamHandler()
_stderr_handler.setFormatter(_stderr_formatter)
_LOGGER.addHandler(_stderr_handler)


def import_module(name):
    """ Imports a module into the current runtime environment

    This function emulates the Python import system that allows for
    importing full path modules.  It will break down the module and
    import each part (or skip if it is already loaded in cache).

    Args:
        name (str): The name of the module to import.  This should be
            the full path of the module

    Returns:
        The module that was imported

    """
    if name in sys.modules:
        # Be sure not to reload a previously loaded module
        mod = sys.modules[name]
    else:
        mod = importlib.import_module(name)
    return mod


def load_module(name):
    """ Attempts to load a module into the current environment

    This function will load a module specified by name.  The module
    name is first checked to see if it is already loaded and will return
    the module if it is.   If the module hasn't been previously loaded
    it will attempt to import it

    Args:
        name (str): Specifies the full name of the module.  For instance
            pyeapi.api.vlans

    Returns:
        The module that has been imported or retrieved from the sys modules

    """
    try:
        mod = None
        mod = sys.modules[name]
    except KeyError:
        mod = import_module(name)
    finally:
        if not mod:
            raise ImportError('unable to import module %s' % name)
        return mod


class ProxyCall(object):

    def __init__(self, proxy, method):
        self.proxy = proxy
        self.method = method

    def __call__(self, *args, **kwargs):
        return self.proxy(self.method, *args, **kwargs)


def islocalconnection():
    """ Checks if running locally on EOS device or remotely

    This function will return a boolean indicating if the current
    execution environment is running locally on an EOS device (True) or
    running remotely and communicating over HTTP/S (False)

    Returns:
        A boolean value that indicates whether or not the current
            thread is local or remote
    """
    return os.path.exists('/etc/Eos-release')


def debug(text):
    """Log a message to syslog and stderr

    Args:
        text (str): The string object to print

    """
    frame = inspect.currentframe().f_back
    module = frame.f_globals['__name__']
    func = frame.f_code.co_name
    msg = "%s.%s: %s" % (module, func, text)
    _LOGGER.debug(msg)


def make_iterable(value):
    """Converts the supplied value to a list object

    This function will inspect the supplied value and return an
    iterable in the form of a list.

    Args:
        value (object): A valid Python object

    Returns:
        An iterable object of type list
    """
    if isinstance(value, str) or isinstance(
            value, dict) or isinstance(value, CliVariants):
        value = [value]

    if not isinstance(value, Iterable):
        raise TypeError('value must be an iterable object')

    return value


def lookahead(it):
    it1, it2 = tee(iter(it))
    next(it2)
    return zip_longest(it1, it2)


def expand_range(arg, value_delimiter=',', range_delimiter='-'):
    """
    Expands a delimited string of ranged integers into a list of strings

    :param arg: The string range to expand
    :param value_delimiter: The delimiter that separates values
    :param range_delimiter: The delimiter that signifies a range of values

    :return: An array of expanded string values
    :rtype: list
    """
    values = list()
    expanded = arg.split(value_delimiter)
    for item in expanded:
        if range_delimiter in item:
            start, end = item.split(range_delimiter)
            _expand = range(int(start), int(end) + 1)
            values.extend([str(x) for x in _expand])
        else:
            values.extend([item])
    return [str(x) for x in values]


def collapse_range(arg, value_delimiter=',', range_delimiter='-'):
    """
    Collapses a list of values into a range set

    :param arg: The list of values to collapse
    :param value_delimiter: The delimiter that separates values
    :param range_delimiter: The delimiter that separates a value range

    :return: An array of collapsed string values
    :rtype: list
    """
    values = list()
    expanded = arg.split(value_delimiter)
    range_start = None

    for v1, v2 in lookahead(expanded):
        if v2:
            v1 = int(v1)
            v2 = int(v2)
            if (v1 + 1) == v2:
                if not range_start:
                    range_start = v1
            elif range_start:
                item = '{}{}{}'.format(range_start, range_delimiter, v1)
                values.extend([item])
                range_start = None
            else:
                values.extend([v1])
        elif range_start:
            item = '{}{}{}'.format(range_start, range_delimiter, v1)
            values.extend([item])
            range_start = None
        else:
            values.extend([v1])
    return [str(x) for x in values]


class CliVariants:
    """
    Provides an interface for cli variants (typically to handle a transition
    period for a deprecated cli)

    Instance must be initialized either with 2 or more str variants:
        ``CliVariants( 'new cli', 'legacy cli' )``,
    or with 2 or more sequences of cli (or a mix of list and str types), e.g.:
        ``CliVariants( ['new cli1', 'new cli2'], 'alt cli3', 'legacy cli4' )``
    """
    @staticmethod
    def expand( cmds ):
        """ Expands cmds argument into a list of all CLI variants

        The method returns a list of all full variant combinations present
        in the the cmds arguement

        Args:
            cmds (list): a list made of str and CliVariants types

        Returns:
            expanded list, e.g.:
                expand( 'x', CliVariants( 'a', 'b'), 'y' )
                will return: [ ['x', 'a', 'y'], ['x', 'b', 'y'] ]
        """
        assert isinstance(cmds, list), 'argument cmds must be list type'
        if not cmds:
            return [ [] ]
        head = cmds[0]
        tail = cmds[1:]
        if isinstance( head, CliVariants ):
            return [ v + e for v in head.variants
                for e in CliVariants.expand( tail ) ]
        else:
            return [ [head] + e for e in CliVariants.expand(tail) ]

    def __init__(self, *cli):
        assert len( cli ) >= 2, 'must be initialized with 2 or more arguments'
        self.variants = [ v if not isinstance(v,
            str) and isinstance(v, Iterable) else [v] for v in cli ]


def _interpolate_docstr( *tkns ):
    """Docstring decorator.
    SYNOPSIS:

         MIN_MTU=68
         MAX_MTU=65535

         @_interpolate_docstr( 'MIN_MTU', 'MAX_MTU' )
         def mtu_check( val ):
            "check mtu against its min value (MIN_MTU) and max value (MAX_MTU)"
            ...

         print( mtu_check.__doc__ )
         check mtu against its min value (68) and max value (65535)
    """
    def docstr_decorator( user_fn ):
        """update user_fn_wrapper doc string with the interpolated user_fn's
        """
        def user_fn_wrapper( *args, **kwargs ):
            return user_fn( *args, **kwargs )
        module = sys.modules[ user_fn.__module__ ]
        docstr = user_fn.__doc__
        for tkn in tkns:
            sval = str( getattr(module, tkn) )
            docstr = docstr.replace( tkn, sval )
        user_fn_wrapper.__doc__ = docstr
        return user_fn_wrapper
    return docstr_decorator



#
# Copyright (c) 2015, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Module for working with EOS MLAG resources

The Mlag resource provides configuration management of global and interface
MLAG settings on an EOS node.

Parameters:

    config (dict): The global MLAG configuration values

    interfaces (dict): The configured MLAG interfaces

Config Parameters:
    domain_id (str): The domain_id parameter  is parsed from the nodes
        mlag configuration.  The domain id is an alphanumeric string that
        names the MLAG domain

    local_interface (str): The local VLAN interface used as the control
        plane endpoint between MLAG peers.  Valid values include any
        VLAN SVI

    peer_address (str): The IP address of the MLAG peer used to send MLAG
        control traffic.  The peer address must be reachable from the
        local interface.  Valid values include any IPv4 unicast address

    peer_link (str): The physical link that connects the node to its MLAG
        peer.  Valid values for the peer link include layer 2 Ethernet or
        Port-Channel interfaces

    shutdown (bool): The administrative state of the global MLAG process.

Interface Parameters:
    mlag_id (str): The interface mlag parameter parsed from the nodes
        interface configuration.  Valid values for the mlag id are in
        the range of 1 to 2000

"""

import re

from pyeapi.api import Entity


class Mlag(Entity):
    """The Mlag class provides management of the MLAG configuration

    The Mlag class is derived from Entity and provides an API for working
    with the nodes MLAG configuraiton.
    """

    def get(self):
        """Returns the Mlag configuration as a resource dict

        Returns:
            dict: A dict ojbect containing the Mlag resource attributes.
        """
        resource = dict()
        resource.update(self._parse_config())
        resource.update(self._parse_interfaces())

        return resource

    def _parse_config(self):
        """Parses the mlag global configuration

        Returns:
            dict: A dict object that is intended to be merged into the
                resource dict
        """
        config = self.get_block('mlag configuration')
        cfg = dict()
        cfg.update(self._parse_domain_id(config))
        cfg.update(self._parse_local_interface(config))
        cfg.update(self._parse_peer_address(config))
        cfg.update(self._parse_peer_link(config))
        cfg.update(self._parse_shutdown(config))
        return dict(config=cfg)

    def _parse_domain_id(self, config):
        """Scans the config block and parses the domain-id value

        Args:
            config (str): The config block to scan

        Returns:
            dict: A dict object that is intended to be merged into the
                resource dict
        """
        match = re.search(r'domain-id (.+)$', config)
        value = match.group(1) if match else None
        return dict(domain_id=value)

    def _parse_local_interface(self, config):
        """Scans the config block and parses the local-interface value

        Args:
            config (str): The config block to scan

        Returns:
            dict: A dict object that is intended to be merged into the
                resource dict
        """
        match = re.search(r'local-interface (\w+)', config)
        value = match.group(1) if match else None
        return dict(local_interface=value)

    def _parse_peer_address(self, config):
        """Scans the config block and parses the peer-address value

        Args:
            config (str): The config block to scan

        Returns:
            dict: A dict object that is intended to be merged into the
                resource dict
        """
        match = re.search(r'peer-address (\d+\.\d+\.\d+\.\d+)$', config)
        value = match.group(1) if match else None
        return dict(peer_address=value)

    def _parse_peer_link(self, config):
        """Scans the config block and parses the peer-link value

        Args:
            config (str): The config block to scan

        Returns:
            dict: A dict object that is intended to be merged into the
                resource dict
        """
        match = re.search(r'peer-link (\S+)', config)
        value = match.group(1) if match else None
        return dict(peer_link=value)

    def _parse_shutdown(self, config):
        """Scans the config block and parses the shutdown value

        Args:
            config (str): The config block to scan

        Returns:
            dict: A dict object that is intended to be merged into the
                resource dict
        """
        value = 'no shutdown' not in config
        return dict(shutdown=value)

    def _parse_interfaces(self):
        """Scans the global config and returns the configured interfaces

        Returns:
            dict: A dict object that is intended to be merged into the
                resource dict.
        """
        interfaces = dict()
        names = re.findall(r'^interface (Po.+)$', self.config, re.M)
        for name in names:
            config = self.get_block('interface %s' % name)
            match = re.search(r'mlag (\d+)', config)
            if match:
                interfaces[name] = dict(mlag_id=match.group(1))
        return dict(interfaces=interfaces)

    def _configure_mlag(self, string, value, default, disable):
        cfg = self.command_builder(string, value=value, default=default,
                                   disable=disable)
        cmds = ['mlag configuration', cfg]
        return super(Mlag, self).configure(cmds)

    def set_domain_id(self, value=None, default=False, disable=False):
        """Configures the mlag domain-id value

        Args:
            value (str): The value to configure the domain-id
            default (bool): Configures the domain-id using the default keyword
            disable (bool): Negates the domain-id using the no keyword

        Returns:
            bool: Returns True if the commands complete successfully
        """
        return self._configure_mlag('domain-id', value, default, disable)

    def set_local_interface(self, value=None, default=False, disable=False):
        """Configures the mlag local-interface value

        Args:
            value (str): The value to configure the local-interface
            default (bool): Configures the local-interface using the
                default keyword
            disable (bool): Negates the local-interface using the no keyword

        Returns:
            bool: Returns True if the commands complete successfully
        """
        return self._configure_mlag('local-interface', value, default, disable)

    def set_peer_address(self, value=None, default=False, disable=False):
        """Configures the mlag peer-address value

        Args:
            value (str): The value to configure the peer-address
            default (bool): Configures the peer-address using the
                default keyword
            disable (bool): Negates the peer-address using the no keyword

        Returns:
            bool: Returns True if the commands complete successfully
        """
        return self._configure_mlag('peer-address', value, default, disable)

    def set_peer_link(self, value=None, default=False, disable=False):
        """Configures the mlag peer-link value

        Args:
            value (str): The value to configure the peer-link
            default (bool): Configures the peer-link using the
                default keyword
            disable (bool): Negates the peer-link using the no keyword

        Returns:
            bool: Returns True if the commands complete successfully
        """
        return self._configure_mlag('peer-link', value, default, disable)

    def set_shutdown(self, default=False, disable=True):
        """Configures the mlag shutdown value

        Default setting for set_shutdown is disable=True, meaning
        'no shutdown'. Setting both default and disable to False will
        effectively enable shutdown.

        Args:
            default (bool): Configures the shutdown using the
                default keyword
            disable (bool): Negates shutdown using the no keyword

        Returns:
            bool: Returns True if the commands complete successfully
        """
        return self._configure_mlag('shutdown', True, default, disable)

    def set_mlag_id(self, name, value=None, default=False, disable=False):
        """Configures the interface mlag value for the specified interface

        Args:
            name (str): The interface to configure.  Valid values for the
                name arg include Port-Channel*
            value (str): The mlag identifier to cofigure on the interface
            default (bool): Configures the interface mlag value using the
                default keyword
            disable (bool): Negates the interface mlag value using the
                no keyword

        Returns:
            bool: Returns True if the commands complete successfully
        """
        cmd = self.command_builder('mlag', value=value, default=default,
                                   disable=disable)
        return self.configure_interface(name, cmd)


def instance(node):
    """Returns an instance of Mlag

    Args:
        node (Node): The node argument passes an instance of Node to the
            resource

    Returns:
        object: An instance of Mlag
    """
    return Mlag(node)



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""API module for Bgp
"""

import re

from collections import namedtuple

import netaddr

from pyeapi.api import Entity, EntityCollection
from pyeapi.utils import make_iterable

Network = namedtuple('Network', 'prefix length route_map')


class Bgp(Entity):
    """The Bgp class implements global BGP router configuration
    """

    def __init__(self, *args, **kwargs):
        super(Bgp, self).__init__(*args, **kwargs)
        self._neighbors = None

    @property
    def neighbors(self):
        if self._neighbors is not None:
            return self._neighbors
        self._neighbors = BgpNeighbors(self.node)
        return self._neighbors

    def get(self):
        """Returns the bgp routing configuration as a dict object
        """
        config = self.get_block('^router bgp .*')
        if not config:
            return None

        response = dict()
        response.update(self._parse_bgp_as(config))
        response.update(self._parse_router_id(config))
        response.update(self._parse_max_paths(config))
        response.update(self._parse_shutdown(config))
        response.update(self._parse_networks(config))

        response['neighbors'] = self.neighbors.getall()

        return response

    def _parse_bgp_as(self, config):
        as_num = re.search(r'(?<=^router bgp ).*', config).group(0)
        return { 'bgp_as': int(as_num) if as_num.isnumeric() else as_num }

    def _parse_router_id(self, config):
        match = re.search(r'router-id ([^\s]+)', config)
        value = match.group(1) if match else None
        return dict(router_id=value)

    def _parse_max_paths(self, config):
        match = re.search(r'maximum-paths\s+(\d+)\s+ecmp\s+(\d+)', config)
        paths = int(match.group(1)) if match else None
        ecmp_paths = int(match.group(2)) if match else None
        return dict(maximum_paths=paths, maximum_ecmp_paths=ecmp_paths)

    def _parse_shutdown(self, config):
        value = 'no shutdown' in config
        return dict(shutdown=not value)

    def _parse_networks(self, config):
        networks = list()
        regexp = r'network (.+)/(\d+)(?: route-map (\w+))*'
        matches = re.findall(regexp, config)
        for (prefix, mask, rmap) in matches:
            rmap = None if rmap == '' else rmap
            networks.append(dict(prefix=prefix, masklen=mask, route_map=rmap))
        return dict(networks=networks)

    def configure_bgp(self, cmd):
        config = self.get()
        cmds = ['router bgp {}'.format(config['bgp_as'])]
        cmds.extend(make_iterable(cmd))
        return super(Bgp, self).configure(cmds)

    def create(self, bgp_as):
        value = int(bgp_as)
        if not 0 < value < 65536:
            raise ValueError('bgp as must be between 1 and 65535')
        command = 'router bgp {}'.format(bgp_as)
        return self.configure(command)

    def delete(self):
        config = self.get()
        if not config:
            return True
        command = 'no router bgp {}'.format(config['bgp_as'])
        return self.configure(command)

    def default(self):
        config = self.get()
        if not config:
            return True
        command = 'default router bgp {}'.format(config['bgp_as'])
        return self.configure(command)

    def set_router_id(self, value=None, default=False, disable=False):
        cmd = self.command_builder('router-id', value=value,
                                   default=default, disable=disable)
        return self.configure_bgp(cmd)

    def set_maximum_paths(self, max_path=None, max_ecmp_path=None,
                          default=False, disable=False):
        if not max_path and max_ecmp_path:
            raise TypeError('Cannot use maximum_ecmp_paths without '
                            'providing max_path')
        value = None
        if max_path:
            value = '{}'.format(max_path)
            if max_ecmp_path:
                value += ' ecmp {}'.format(max_ecmp_path)
        cmd = self.command_builder('maximum-paths', value=value,
                                   default=default, disable=disable)
        return self.configure_bgp(cmd)

    def set_shutdown(self, default=False, disable=True):
        # Default setting for BGP shutdown is disable=True,
        # meaning 'no shutdown'.
        # If both default and disable are false, BGP shutdown will
        # effectively be enabled.
        cmd = self.command_builder('shutdown', value=True, default=default,
                                   disable=disable)
        return self.configure_bgp(cmd)

    def add_network(self, prefix, length, route_map=None):
        if prefix == '' or length == '':
            raise ValueError('network prefix and length values '
                             'may not be empty')
        cmd = 'network {}/{}'.format(prefix, length)
        if route_map:
            cmd += ' route-map {}'.format(route_map)
        return self.configure_bgp(cmd)

    def remove_network(self, prefix, masklen, route_map=None):
        if prefix == '' or masklen == '':
            raise ValueError('network prefix and length values '
                             'may not be empty')
        cmd = 'no network {}/{}'.format(prefix, masklen)
        if route_map:
            cmd += ' route-map {}'.format(route_map)
        return self.configure_bgp(cmd)


class BgpNeighbors(EntityCollection):

    def get(self, name):
        config = self.get_block('^router bgp .*')
        response = dict(name=name)
        response.update(self._parse_peer_group(config, name))
        response.update(self._parse_remote_as(config, name))
        response.update(self._parse_send_community(config, name))
        response.update(self._parse_shutdown(config, name))
        response.update(self._parse_description(config, name))
        response.update(self._parse_next_hop_self(config, name))
        response.update(self._parse_route_map_in(config, name))
        response.update(self._parse_route_map_out(config, name))
        return response

    def getall(self):
        config = self.get_block('^router bgp .*')
        if not config:
            return None

        collection = dict()
        for neighbor in re.findall(r'neighbor ([^\s]+)', config):
            collection[neighbor] = self.get(neighbor)
        return collection

    def _parse_peer_group(self, config, name):
        if self.version_number >= '4.23':
            regexp = r'neighbor {} peer group ([^\s]+)'.format(name)
        else:
            regexp = r'neighbor {} peer-group ([^\s]+)'.format(name)
        match = re.search(regexp, config)
        value = match.group(1) if match else None
        return dict(peer_group=value)

    def _parse_remote_as(self, config, name):
        remote_as_re = rf'(?<=neighbor {name} remote-as ).*'
        match = re.search(remote_as_re, config)
        return { 'remote_as': match.group(0) if match else None}

    def _parse_send_community(self, config, name):
        exp = 'no neighbor {} send-community'.format(name)
        value = exp in config
        return dict(send_community=not value)

    def _parse_shutdown(self, config, name):
        regexp = r'(?<!no )neighbor {} shutdown'.format(name)
        match = re.search(regexp, config, re.M)
        value = True if match else False
        return dict(shutdown=value)

    def _parse_description(self, config, name):
        regexp = r'neighbor {} description (.*)$'.format(name)
        match = re.search(regexp, config, re.M)
        value = match.group(1) if match else None
        return dict(description=value)

    def _parse_next_hop_self(self, config, name):
        exp = 'no neighbor {} next-hop-self'.format(name)
        value = exp in config
        return dict(next_hop_self=not value)

    def _parse_route_map_in(self, config, name):
        regexp = r'neighbor {} route-map ([^\s]+) in'.format(name)
        match = re.search(regexp, config, re.M)
        value = match.group(1) if match else None
        return dict(route_map_in=value)

    def _parse_route_map_out(self, config, name):
        regexp = r'neighbor {} route-map ([^\s]+) out'.format(name)
        match = re.search(regexp, config, re.M)
        value = match.group(1) if match else None
        return dict(route_map_out=value)

    def ispeergroup(self, name):
        try:
            netaddr.IPAddress(name)
            return False
        except netaddr.core.AddrFormatError:
            return True

    def create(self, name):
        return self.set_shutdown(name, default=False, disable=False)

    def delete(self, name):
        response = self.configure('no neighbor {}'.format(name))
        if not response:
            if self.version_number >= '4.23':
                response = self.configure('no neighbor {} '
                                          'peer group'.format(name))
            else:
                response = self.configure('no neighbor {} '
                                          'peer-group'.format(name))
        return response

    def configure(self, cmd):
        match = re.search(r'router bgp (\d+)', self.config)
        if not match:
            raise ValueError('bgp is not configured')
        cmds = ['router bgp {}'.format(match.group(1)), cmd]
        return super(BgpNeighbors, self).configure(cmds)

    def command_builder(self, name, cmd, value, default, disable):
        string = 'neighbor {} {}'.format(name, cmd)
        return super(BgpNeighbors, self).command_builder(string, value,
                                                         default, disable)

    def set_peer_group(self, name, value=None, default=False, disable=False):
        if not self.ispeergroup(name):
            if self.version_number >= '4.23':
                cmd = self.command_builder(name, 'peer group', value, default,
                                           disable)
            else:
                cmd = self.command_builder(name, 'peer-group', value, default,
                                           disable)
            return self.configure(cmd)
        return False

    def set_remote_as(self, name, value=None, default=False, disable=False):
        cmd = self.command_builder(name, 'remote-as', value, default, disable)
        return self.configure(cmd)

    def set_shutdown(self, name, default=False, disable=True):
        # Default setting for BGP neighbor shutdown is
        # disable=True, meaning 'no shutdown'
        # If both default and disable are false, BGP neighbor shutdown will
        # effectively be enabled.
        cmd = self.command_builder(name, 'shutdown', True, default, disable)
        return self.configure(cmd)

    def set_send_community(self, name, value=None, default=False,
                           disable=False):
        cmd = self.command_builder(name, 'send-community', value, default,
                                   disable)
        return self.configure(cmd)

    def set_next_hop_self(self, name, value=None, default=False,
                          disable=False):
        cmd = self.command_builder(name, 'next-hop-self', value, default,
                                   disable)
        return self.configure(cmd)

    def set_route_map_in(self, name, value=None, default=False, disable=False):
        cmd = self.command_builder(name, 'route-map', value, default, disable)
        cmd += ' in'
        return self.configure(cmd)

    def set_route_map_out(self, name, value=None, default=False,
                          disable=False):
        cmd = self.command_builder(name, 'route-map', value, default, disable)
        cmd += ' out'
        return self.configure(cmd)

    def set_description(self, name, value=None, default=False, disable=False):
        cmd = self.command_builder(name, 'description', value, default,
                                   disable)
        return self.configure(cmd)


def instance(api):
    return Bgp(api)
